#! /usr/bin/env python

"""
Module with the MCMC (``emcee``) sampling for model spectra parameter estimation.
"""


__author__ = 'V. Christiaens, O. Wertz, Carlos Alberto Gomez Gonzalez'
__all__ = ['mcmc_spec_sampling',
           'spec_lnprob',
           'spec_chain_zero_truncated',
           'spec_show_corner_plot',
           'spec_show_walk_plot',
           'spec_confidence']

from astropy import constants as con
import numpy as np
import os
from os.path import isdir, isfile
import emcee
import inspect
import datetime
import corner
from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
import pickle
from scipy.stats import norm
from ..conf import time_ini, timing
from ..conf.utils_conf import sep
from ..fits import open_fits, write_fits
from ..negfc.utils_mcmc import gelman_rubin, autocorr_test
from .chi import goodness_of_fit
from .model_resampling import make_model_from_params, make_resampled_models
from .utils_spec import mj_from_rj_and_logg
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def spec_lnprior(params, labels, bounds, priors=None):
    """ Define the prior log-function.
    
    Parameters
    ----------
    params: tuple
        The model parameters.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
    bounds: dictionary
        Each entry should be associated with a tuple corresponding to lower and 
        upper bounds respectively. Bounds should be provided for ALL model
        parameters, including 'R' (planet photometric radius). 'Av' (optical 
        extinction) is optional. If provided here, Av will also be fitted.
        All keywords that are neither 'R', 'Av' nor 'M' will 
        be considered model grid parameters.
        Example for BT-SETTL: bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5),
        'R':(0.1,5), 'Av':(0.,2.5)}
        'M' can be used for a prior on the mass of the planet. In that case the
        corresponding prior log probability is computed from the values for 
        parameters 'logg' and 'R'.
    priors: dictionary, opt
        If not None, sets prior estimates for each parameter of the model. Each 
        entry should be set to either None (no prior) or a tuple of 2 elements 
        containing prior estimate and uncertainty on the estimate.
        Missing entries (i.e. provided in bounds dictionary but not here) will
        be associated no prior.
        e.g. priors = {'Teff':(1600,100), 'logg':(3.5,0.5),
                       'R':(1.6,0.1), 'Av':(1.8,0.2), 'M':(10,3)}
        Important: dictionary entry names should match exactly those of bounds.
    
    Returns
    -------
    out: float.
        If all parameters are within bounds:
            * 0 if no prior,
            * the sum of gaussian log proba for each prior otherwise.
        If at least one model parameters is out of bounds:
        returns -np.inf 
    """
    n_params = len(params)
    n_labs = len(labels)
    n_dico = len(bounds)    
    if n_dico!=n_params or n_dico != n_labs:
        raise TypeError('params, labels and bounds should have same length')

    cond = True
    for pp in range(n_params):
        if not bounds[labels[pp]][0] <= params[pp] <= bounds[labels[pp]][1]:
            cond=False

    if cond:
        lnprior = 0.
        if priors is not None:
            for key, prior in priors.items():
                if key == 'M' and 'logg' in labels:
                    idx_logg =labels.index("logg")
                    idx_rp = labels.index("R")
                    rp = params[idx_rp]
                    logg = params[idx_logg]
                    mp = mj_from_rj_and_logg(rp, logg)
                    lnprior += -0.5 * (mp - prior[0])**2 / prior[1]**2
                else:
                    idx_prior = labels.index(key)
                    lnprior += -0.5*(params[idx_prior]-prior[0])**2/prior[1]**2
        return lnprior
    else:
        return -np.inf


def spec_lnlike(params, labels, grid_param_list, lbda_obs, spec_obs, err_obs, 
                dist, model_grid=None, model_reader=None, em_lines={}, 
                em_grid={}, dlbda_obs=None, instru_corr=None, 
                instru_fwhm=None, instru_idx=None, filter_reader=None, 
                AV_bef_bb=False, units_obs='si', units_mod='si', interp_order=1):
    """ Define the likelihood log-function.
    
    Parameters
    ----------
    params : tuple
        Set of models parameters for which the model grid has to be 
        interpolated.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
    grid_param_list : list of 1d numpy arrays/lists OR None
        - If list, should contain list/numpy 1d arrays with available grid of 
        model parameters. 
        - Set to None for a pure n-blackbody fit, n=1,2,...
        - Note1: model grids should not contain grids on radius and Av, but 
        these should still be passed in initial_state (Av optional).
        - Note2: for a combined grid model + black body, just provide
        the grid parameter list here, and provide values for 'Tbbn' and 'Rbbn'
        in initial_state, labels and bounds.
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    model_grid : numpy N-d array, optional
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of n_T values of Teff 
        and n_g values of Logg, the numpy array should be n_T x n_g x n_ch x 2, 
        where n_ch is the number of wavelengths for the observed spectrum,
        and the last 2 dims are for wavelength and fluxes respectively.
        If provided, takes precedence over model_name/model_reader.
    model_reader : python routine, opt
        External routine that reads a model file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains model values. See example routine in model_interpolation() 
        description.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        4 values: 
            1) the wavelength (in mu); 
            2) a string indicating whether line intensity is expressed in flux 
            ('F'), luminosity ('L') or log(L/LSun) ("LogL");
            3) the FWHM of the gaussian (or None if to be set automatically); 
            4) whether the FWHM is expressed in 'nm', 'mu' or 'km/s'. 
        The third and fourth can also be set to None. In that case, the FWHM of 
        the gaussian will automatically be set to the equivalent width of the
        line, calculated from the flux to be injected and the continuum 
        level (measured in the grid model to which the line is injected). 
        Examples: em_lines = {'BrG':(2.1667,'F',None, None)};
                  em_lines = {'BrG':(2.1667,'LogL', 100, 'km/s')}
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    AV_bef_bb: bool, optional
        If both extinction and an extra bb component are free parameters, 
        whether to apply extinction before adding the BB component (e.g. 
        extinction mostly from circumplanetary dust) or after the BB component
        (e.g. mostly insterstellar extinction).
    units_obs : str, opt {'si','cgs','jy'}
        Units of observed spectrum. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu 
        or 'jy' for janskys.
    units_mod: str, opt {'si','cgs','jy'}
        Units of the model. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu or 'jy'
        for janskys. If different to units_obs, the spectrum units will be 
        converted.
    interp_order: int, opt, {-1,0,1} 
        Interpolation mode for model interpolation.
        -1: log interpolation (i.e. linear interpolatlion on log(Flux))
        0: nearest neighbour model.
        1: Order 1 spline interpolation.
        
    Returns
    -------
    out: float
        The log of the likelihood.
        
    """

    if grid_param_list is not None:
        if model_grid is None and model_reader is None:
            msg = "model_name and model_reader must be provided"
            raise TypeError(msg)
                           
    lbda_mod, spec_mod = make_model_from_params(params, labels, grid_param_list, 
                                                dist, lbda_obs, model_grid, 
                                                model_reader, em_lines, em_grid, 
                                                dlbda_obs, instru_fwhm, 
                                                instru_idx, filter_reader, 
                                                AV_bef_bb, units_obs, units_mod, 
                                                interp_order)
    
    # evaluate the goodness of fit indicator
    chi = goodness_of_fit(lbda_obs, spec_obs, err_obs, lbda_mod, spec_mod, 
                          dlbda_obs=dlbda_obs, instru_corr=instru_corr, 
                          instru_fwhm=instru_fwhm, instru_idx=instru_idx, 
                          filter_reader=filter_reader, plot=False, outfile=None)
    
    # log likelihood
    lnlikelihood = -0.5 * chi
    
    return lnlikelihood


def spec_lnprob(params, labels, bounds, grid_param_list, lbda_obs, spec_obs, 
                err_obs, dist, model_grid=None, model_reader=None, em_lines={},
                em_grid={}, dlbda_obs=None, instru_corr=None, instru_fwhm=None, 
                instru_idx=None, filter_reader=None, AV_bef_bb=False, 
                units_obs='si', units_mod='si', interp_order=1, priors=None, 
                physical=True):
    """ Define the probability log-function as the sum between the prior and
    likelihood log-functions.
    
    Parameters
    ----------
    params: tuple
        The model parameters.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
    bounds: dictionary
        Each entry should be associated with a tuple corresponding to lower and 
        upper bounds respectively. Bounds should be provided for ALL model
        parameters, including 'R' (planet photometric radius). 'Av' (optical 
        extinction) is optional. If provided here, Av will also be fitted.
        All keywords that are neither 'R', 'Av' nor 'M' will 
        be considered model grid parameters.
        Example for BT-SETTL: bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5),
        'R':(0.1,5), 'Av':(0.,2.5)}
        'M' can be used for a prior on the mass of the planet. In that case the
        corresponding prior log probability is computed from the values for 
        parameters 'logg' and 'R'.
    grid_param_list : list of 1d numpy arrays/lists OR None
        - If list, should contain list/numpy 1d arrays with available grid of 
        model parameters. 
        - Set to None for a pure n-blackbody fit, n=1,2,...
        - Note1: model grids should not contain grids on radius and Av, but 
        these should still be passed in initial_state (Av optional).
        - Note2: for a combined grid model + black body, just provide
        the grid parameter list here, and provide values for 'Tbbn' and 'Rbbn'
        in initial_state, labels and bounds.
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    model_grid : numpy N-d array, optional
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of n_T values of Teff 
        and n_g values of Logg, the numpy array should be n_T x n_g x n_ch x 2, 
        where n_ch is the number of wavelengths for the observed spectrum,
        and the last 2 dims are for wavelength and fluxes respectively.
        If provided, takes precedence over model_name/model_reader.
    model_reader : python routine
        External routine that reads a model file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains model values. See example routine in model_interpolation() 
        description.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        4 values: 
            1) the wavelength (in mu); 
            2) a string indicating whether line intensity is expressed in flux 
            ('F'), luminosity ('L') or log(L/LSun) ("LogL");
            3) the FWHM of the gaussian (or None if to be set automatically); 
            4) whether the FWHM is expressed in 'nm', 'mu' or 'km/s'. 
        The third and fourth can also be set to None. In that case, the FWHM of 
        the gaussian will automatically be set to the equivalent width of the
        line, calculated from the flux to be injected and the continuum 
        level (measured in the grid model to which the line is injected). 
        Examples: em_lines = {'BrG':(2.1667,'F',None, None)};
                  em_lines = {'BrG':(2.1667,'LogL', 100, 'km/s')}
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    AV_bef_bb: bool, optional
        If both extinction and an extra bb component are free parameters, 
        whether to apply extinction before adding the BB component (e.g. 
        extinction mostly from circumplanetary dust) or after the BB component
        (e.g. mostly insterstellar extinction).
    units_obs : str, opt {'si','cgs','jy'}
        Units of observed spectrum. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu 
        or 'jy' for janskys.
    units_mod: str, opt {'si','cgs','jy'}
        Units of the model. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu or 'jy'
        for janskys. If different to units_obs, the spectrum units will be 
        converted.
    interp_order: int, opt, {-1,0,1} 
        Interpolation mode for model interpolation.
        -1: log interpolation (i.e. linear interpolatlion on log(Flux))
        0: nearest neighbour model.
        1: Order 1 spline interpolation.
    priors: dictionary, opt
        If not None, sets prior estimates for each parameter of the model. Each 
        entry should be set to either None (no prior) or a tuple of 2 elements 
        containing prior estimate and uncertainty on the estimate.
        Missing entries (i.e. provided in bounds dictionary but not here) will
        be associated no prior.
        e.g. priors = {'Teff':(1600,100), 'logg':(3.5,0.5),
                       'R':(1.6,0.1), 'Av':(1.8,0.2), 'M':(10,3)}
        Important: dictionary entry names should match exactly those of bounds.
    physical: bool, opt
        In case of extra black body component(s) to a photosphere, whether to 
        force lower temperature than the photosphere effective temperature.
        
    Returns
    -------
    out: float
        The probability log-function.
    
    """
    
    lp = spec_lnprior(params, labels, bounds, priors)
    
    if np.isinf(lp):
        return -np.inf
    
    if 'Tbb1' in labels:
        condT = ('T' in labels or 'Teff' in labels)
        n_bb = 0
        for label in labels:
            if 'Tbb' in label:
                n_bb+=1
        idx_Tbb1 = labels.index("Tbb1")
        for ii in range(n_bb):
            idx = ii*2
            if grid_param_list is not None and condT:
                if 'T' in labels:
                    idx_Teff = labels.index("T")
                else:
                    idx_Teff = labels.index("Teff")
                # COND 1: only allow Tbb < Teff
                if physical and params[idx_Tbb1+idx] >= params[idx_Teff]:
                    return -np.inf
                elif physical:
                    idx_R= labels.index("R")
                    R = params[idx_R]
                    Rbb = params[idx_Tbb1+idx+1]
                    if Rbb<=R:
                        pass
                    else:
                        Teq = params[idx_Teff] * np.sqrt(R/(2*Rbb))
                        # COND 2: only allow Tbb <= Teq at Rbb
                        if params[idx_Tbb1+idx] >= Teq:
                            return -np.inf
            if ii>0:
                # avoid unnecessary degenerecaies by only allowing:
                # Teff1 > Teff2, Teff2 > Teff3
                if params[idx_Tbb1+idx] >= params[idx_Tbb1+idx-2]:
                    return -np.inf
        
    return lp + spec_lnlike(params, labels, grid_param_list, lbda_obs, spec_obs, 
                            err_obs, dist, model_grid, model_reader, em_lines,
                            em_grid, dlbda_obs, instru_corr, instru_fwhm, 
                            instru_idx, filter_reader, AV_bef_bb, units_obs, 
                            units_mod, interp_order)


def mcmc_spec_sampling(lbda_obs, spec_obs, err_obs, dist, grid_param_list, 
                       initial_state, labels, bounds, resamp_before=True, 
                       model_grid=None, model_reader=None, em_lines={}, 
                       em_grid={}, dlbda_obs=None, instru_corr=None, 
                       instru_fwhm=None, instru_idx=None, filter_reader=None, 
                       AV_bef_bb=False, units_obs='si', units_mod='si', 
                       interp_order=1, priors=None, physical=True, 
                       interp_nonexist=True, ini_ball=1e-1, a=2.0, 
                       nwalkers=1000, niteration_min=10, niteration_limit=1000, 
                       niteration_supp=0, check_maxgap=20, conv_test='ac', 
                       ac_c=50, ac_count_thr=3, burnin=0.3, rhat_threshold=1.01, 
                       rhat_count_threshold=1, grid_name='resamp_grid.fits', 
                       output_dir='specfit/', output_file=None, nproc=1, 
                       display=False, verbosity=0, save=False):
    """ Runs an affine invariant MCMC sampling algorithm in order to determine
    the most likely parameters for given spectral model and observed spectrum. 
    Allowed features:
        - Spectral models can either be read from a grid (e.g. BT-SETTL) or 
        be purely parametric (e.g. a blackbody model). 
        - Extinction (A_V) and total-to-selective optical extinction ratio (R_V)
        can be sampled. Default: A_V=0. If non-zero, default R_V=3.1 (ISM).
        - A dictionary of emission lines can be provided and their flux can be 
        sampled too.
        - Gaussian priors can be provided for each parameter, including the 
        mass of the object. The latter will be used if 'logg' is a parameter.
        - Spectral correlation between measurements will be taken into account 
        if provided in 'instru_corr'.
        - Convolution of the model spectra with instrumental FWHM or 
        photometric filter can be performed using 'instru_fwhm' and/or 
        'filter_reader' (done before resampling to observed).
        - The weight of each observed point will be directly proportional to
        Delta lbda_i/lbda_i, where Delta lbda_i is either the FWHM of the 
        photometric filter (imager) or the width of the spectral channel (IFS).
        - MCMC convergence criterion can either be based on auto-correlation 
        time (default) or the Gelman-Rubin test.
    The result of this procedure is a chain with the samples from the posterior 
    distributions of each of the free parameters in the model.
    More details in Christiaens et al. (2021).
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    dist :  float
        Distance in parsec, used for flux scaling of the models.
    grid_param_list : list of 1d numpy arrays/lists OR None
        - If list, should contain list/numpy 1d arrays with available grid of 
        model parameters. 
        - Set to None for a pure n-blackbody fit, n=1,2,...
        - Note1: model grids should not contain grids on radius and Av, but 
        these should still be passed in initial_state (Av optional).
        - Note2: for a combined grid model + black body, just provide
        the grid parameter list here, and provide values for 'Tbbn' and 'Rbbn'
        in initial_state, labels and bounds.
    initial_state: tuple of floats
        Initial guess on the best fit parameters of the spectral fit. Length of 
        the tuple should match the total number of free parameters. Walkers
        will all be initialised in a small ball of parameter space around that
        first guess.
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the intensity of emission lines (labels must match 
        those in the em_lines dict), in units of the model spectrum (x mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - then the planet photometric radius 'R', in Jupiter radius
        - (optionally) the flux of emission lines (labels should match those in
        the em_lines dictionary), in units of the model spectrum (times mu)
        - (optionally) the optical extinction 'Av', in mag
        - (optionally) the ratio of total to selective optical extinction 'Rv'
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution.
    bounds: dictionary
        Each entry should be associated with a tuple corresponding to lower and 
        upper bounds respectively. Bounds should be provided for ALL model
        parameters, including 'R' (planet photometric radius). 'Av' (optical 
        extinction) is optional. If provided here, Av will also be fitted.
        Example for BT-SETTL: bounds = {'Teff':(1000,2000), 'logg':(3.0,4.5),
        'R':(0.1,5), 'Av':(0.,2.5)}
        'M' can be used for a prior on the mass of the planet. In that case the
        corresponding prior log probability is computed from the values for 
        parameters 'logg' and 'R' (if both exist).
    resamp_before: bool, optional
        Whether to prepare the whole grid of resampled models before entering 
        the MCMC, i.e. to avoid doing it at every MCMC step. Recommended.
        Only reason not to: model grid is too large and individual models 
        require being opened and resampled at each step.
    model_grid : numpy N-d array, optional
        If provided, should contain the grid of model spectra for each
        free parameter of the given grid. I.e. for a grid of n_T values of Teff 
        and n_g values of Logg, the numpy array should be n_T x n_g x n_ch x 2, 
        where n_ch is the number of wavelengths for the observed spectrum,
        and the last 2 dims are for wavelength and fluxes respectively.
        If provided, takes precedence over filename/file_reader.
    model_reader : python routine, optional
        External routine that reads a model file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains model values. See example routine in model_interpolation() 
        description.
    em_lines: dictionary, opt
        Dictionary of emission lines to be added on top of the model spectrum.
        Each dict entry should be the name of the line, assigned to a tuple of
        4 values: 
            1) the wavelength (in mu); 
            2) a string indicating whether line intensity is expressed in flux 
            ('F'), luminosity ('L') or log(L/LSun) ("LogL");
            3) the FWHM of the gaussian (or None if to be set automatically); 
            4) whether the FWHM is expressed in 'nm', 'mu' or 'km/s'. 
        The third and fourth can also be set to None. In that case, the FWHM of 
        the gaussian will automatically be set to the equivalent width of the
        line, calculated from the flux to be injected and the continuum 
        level (measured in the grid model to which the line is injected). 
        Examples: em_lines = {'BrG':(2.1667,'F',None, None)};
                  em_lines = {'BrG':(2.1667,'LogL', 100, 'km/s')}
    em_grid: dictionary pointing to lists, opt
        Dictionary where each entry corresponds to an emission line and points
        to a list of values to inject for emission line fluxes. For computation 
        efficiency, interpolation will be performed between the points of this 
        grid during the MCMC sampling. Dict entries should match labels and 
        em_lines.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float OR list of either floats or strings, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band filter FWHM if 
        relevant.
        If strings are provided, they should correspond to filenames (including 
        full paths) of text files containing the filter information for each 
        observed wavelength. Strict format: 
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default file reader will be used. 
        It assumes the following format for the files:
        - first row containing header
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header key
        name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer or
        "lambda(nm)" for nanometer. Note: Only what is in parentheses matters.
        Important: filter files should all have the same format and WL units.
    AV_bef_bb: bool, optional
        If both extinction and an extra bb component are free parameters, 
        whether to apply extinction before adding the BB component (e.g. 
        extinction mostly from circumplanetary dust) or after the BB component
        (e.g. mostly insterstellar extinction).
    units_obs : str, opt {'si','cgs','jy'}
        Units of observed spectrum. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu 
        or 'jy' for janskys.
    units_mod: str, opt {'si','cgs','jy'}
        Units of the model. 'si' for W/m^2/mu; 'cgs' for ergs/s/cm^2/mu or 'jy'
        for janskys. If different to units_obs, the spectrum units will be 
        converted.
    interp_order: int, opt, {-1,0,1} 
        Interpolation mode for model interpolation.
        -1: log interpolation (i.e. linear interpolatlion on log(Flux))
        0: nearest neighbour model.
        1: Order 1 spline interpolation.
    priors: dictionary, opt
        If not None, sets prior estimates for each parameter of the model. Each 
        entry should be set to either None (no prior) or a tuple of 2 elements 
        containing prior estimate and uncertainty on the estimate.
        Missing entries (i.e. provided in bounds dictionary but not here) will
        be associated no prior.
        e.g. priors = {'Teff':(1600,100), 'logg':(3.5,0.5),
                       'R':(1.6,0.1), 'Av':(1.8,0.2), 'M':(10,3)}
        Important: dictionary entry names should match exactly those of bounds.
    physical: bool, opt
        In case of extra black body component(s) to a photosphere, whether to 
        force lower temperature than the photosphere effective temperature.
    interp_nonexist: bool, opt
        Whether to interpolate non-existing models in the grid. Only used if 
        resamp_before is set to True.
    ini_ball: float or string, default=1e-1
        Size of the initial ball in parameter space from which walkers start 
        their chain. If "uniform" is provided, a uniform ini_ball spanning
        the bounds interval will be used to initialise walkers.
    a: float, default=2.0
        The proposal scale parameter. See notes.
    nwalkers: int, default: 1000
        Number of walkers
    niteration_min: int, optional
        Steps per walker lower bound. The simulation will run at least this
        number of steps per walker.
    niteration_limit: int, optional
        Steps per walker upper bound. If the simulation runs up to
        'niteration_limit' steps without having reached the convergence
        criterion, the run is stopped.
    niteration_supp: int, optional
        Number of iterations to run after having "reached the convergence".
    burnin: float, default=0.3
        The fraction of a walker which is discarded.
    rhat_threshold: float, default=0.01
        The Gelman-Rubin threshold used for the test for nonconvergence.
    rhat_count_threshold: int, optional
        The Gelman-Rubin test must be satisfied 'rhat_count_threshold' times in
        a row before claiming that the chain has converged.
    check_maxgap: int, optional
        Maximum number of steps per walker between two convergence tests.
    conv_test: str, optional {'gb','autocorr'}
        Method to check for convergence: 
        - 'gb' for gelman-rubin test
        (http://digitalassets.lib.berkeley.edu/sdtr/ucb/text/305.pdf)
        - 'autocorr' for autocorrelation analysis 
        (https://emcee.readthedocs.io/en/stable/tutorials/autocorr/)
    nproc: int, optional
        The number of processes to use for parallelization.
    grid_name: str, optional
        Name of the fits file containing the model grid (numpy array) AFTER
        convolution+resampling as the observed spectrum given as input.
        If provided, will read it if it exists (and resamp_before is set
        to True), or make it and write it if it doesn't.
    output_dir: str, optional
        The name of the output directory which contains the output files in the 
        case  ``save`` is True.        
    output_file: str, optional
        The name of the output file which contains the MCMC results in the case
        ``save`` is True.
    display: bool, optional
        If True, the walk plot is displayed at each evaluation of the Gelman-
        Rubin test.
    verbosity: 0, 1 or 2, optional
        Verbosity level. 0 for no output and 2 for full information.
    save: bool, optional
        If True, the MCMC results are pickled.
                    
    Returns
    -------
    out : numpy.array
        The MCMC samples after truncation of zeros.
    lnprobability: emcee sample object
        The corresponding probabilities for each sample
        
    Notes
    -----
    The parameter `a` must be > 1. For more theoretical information concerning
    this parameter, see Goodman & Weare, 2010, Comm. App. Math. Comp. Sci.,
    5, 65, Eq. [9] p70.
    
    The parameter 'rhat_threshold' can be a numpy.array with individual
    threshold value for each model parameter.
    """
    nparams = len(initial_state)
    
    if grid_param_list is not None:
        if model_grid is None and model_reader is None:
            msg = "Either model_grid or model_reader have to be provided"
            raise TypeError(msg)
        n_gparams = len(grid_param_list)
        gp_dims = []
        for nn in range(n_gparams):
            gp_dims.append(len(grid_param_list[nn]))
        gp_dims = tuple(gp_dims)
    else:
        n_gparams = 0
        
    # format emission line dictionary and em grid
    if len(em_grid)>0:
        n_em = len(em_grid)
        em_dims = []
        for lab in labels:
            if lab in em_grid.keys():
                em_dims.append(len(em_grid[lab]))
        em_dims = tuple(em_dims)
        # update the grids depending on input units => make it surface flux
        idx_R = labels.index('R')
        for key, val in em_lines.items():
            if val[1] == 'L':
                idx_line = labels.index(key)
                # adapt grid
                R_si = initial_state[idx_R]*con.R_jup.value
                conv_fac = 4*np.pi*R_si**2
                tmp = np.array(em_grid[key])/conv_fac
                em_grid[key] = tmp.tolist()
                # adapt ini state 
                initial_state = list(initial_state)
                initial_state[idx_line] /= conv_fac
                initial_state = tuple(initial_state)
                #adapt bounds
                bounds_ori = list(bounds[key])
                bounds[key] = (bounds_ori[0]/conv_fac, bounds_ori[1]/conv_fac) 
            elif val[1] == 'LogL':
                idx_line = labels.index(key)
                R_si = initial_state[idx_R]*con.R_jup.value
                conv_fac = con.L_sun.value/(4*np.pi*R_si**2)
                tmp = np.power(10,np.array(em_grid[key]))*conv_fac
                em_grid[key] = tmp.tolist()  
                # adapt ini state 
                initial_state = list(initial_state)
                initial_state[idx_line] = conv_fac*10**initial_state[idx_line]
                initial_state = tuple(initial_state)
                #adapt bounds
                bounds_ori = list(bounds[key])
                bounds[key] = (conv_fac*10**bounds_ori[0], 
                               conv_fac*10**bounds_ori[1]) 
            if em_lines[key][2] is not None:
                if em_lines[key][-1] == 'km/s':
                    v = em_lines[key][2]
                    em_lines_tmp = list(em_lines[key])
                    em_lines_tmp[2] = (1000*v/con.c.value)*em_lines[key][0]
                    em_lines_tmp[3] = 'mu'
                    em_lines[key] = tuple(em_lines_tmp)
                elif em_lines[key][-1] == 'nm':
                    em_lines_tmp = list(em_lines[key])
                    em_lines_tmp[2] = em_lines[key][2]/1000
                    em_lines_tmp[3] = 'mu'
                    em_lines[key] = tuple(em_lines_tmp)
                elif em_lines[key][-1] != 'mu':
                    msg = "Invalid unit of FWHM for line injection"
                    raise ValueError(msg)
                
    if model_grid is not None and grid_param_list is not None:
        if model_grid.ndim-2 != n_gparams:
            msg = "Ndim-2 of model_grid should match len(grid_param_list)"
            raise TypeError(msg)
    
    if verbosity > 0:
        start_time = time_ini()
        print("       MCMC sampler for spectral fitting       ")
        print(sep)


    # If required, create the output folder.
    if save or resamp_before:
        if not output_dir:
            raise ValueError('Please provide an output directory path')
        if not isdir(output_dir):
            os.makedirs(output_dir)
        if output_dir[-1] != '/':
            output_dir = output_dir+'/'

        #output_file_tmp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
 
    # Check model grid parameters extend beyond bounds to avoid extrapolation
    if grid_param_list is not None:
        for pp in range(n_gparams):
            if grid_param_list[pp][0]>bounds[labels[pp]][0]:
                msg= "Grid has to extend beyond bounds for {}."
                msg+="\n Consider increasing the lower bound to >{}."
                raise ValueError(msg.format(labels[pp],grid_param_list[pp][0]))
            if grid_param_list[pp][-1]<bounds[labels[pp]][1]:
                msg= "Grid has to extend beyond bounds for {}."
                msg+="\n Consider decreasing the upper bound to <{}."
                raise ValueError(msg.format(labels[pp],grid_param_list[pp][1]))
                
    # Check initial state is within bounds for all params (not only grid)
    for pp in range(nparams):
        if initial_state[pp]<bounds[labels[pp]][0]:
            msg= "Initial state has to be within bounds for {}."
            msg+="\n Consider decreasing the lower bound to <{}."
            raise ValueError(msg.format(labels[pp],initial_state[pp]))            
        if initial_state[pp]>bounds[labels[pp]][1]:
            msg= "Initial state has to be within bounds for {}."
            msg+="\n Consider decreasing the upper bound to >{}."
            raise ValueError(msg.format(labels[pp],initial_state[pp]))
        
    # Prepare model grid: convolve+resample models as observations 
    if resamp_before and grid_param_list is not None:
        if isfile(output_dir+grid_name):
            model_grid = open_fits(output_dir+grid_name)
            # check its shape is consistent with grid_param_list
            if model_grid.shape[:n_gparams] != gp_dims:
                msg="the loaded model grid ({}) doesn't have expected dims ({})"
                raise TypeError(msg.format(model_grid.shape,gp_dims))
            elif model_grid.shape[-2] != len(lbda_obs):
                msg="the loaded model grid doesn't have expected WL dimension"
                raise TypeError(msg)
            elif model_grid.shape[-1] != 2:
                msg="the loaded model grid doesn't have expected last dimension"
                raise TypeError(msg)
            elif len(em_grid) > 0:
                if model_grid.shape[n_gparams:n_gparams+n_em] != em_dims:
                    msg="loaded model grid ({}) doesn't have expected dims ({})"
                    raise TypeError(msg.format(model_grid.shape,em_dims))
        else:
            model_grid = make_resampled_models(lbda_obs, grid_param_list, 
                                               model_grid, model_reader, 
                                               em_lines, em_grid, dlbda_obs, 
                                               instru_fwhm, instru_idx, 
                                               filter_reader, interp_nonexist)
            if output_dir and grid_name:
                write_fits(output_dir+grid_name, model_grid)
        # note: if model_grid is provided, it is still resampled to the 
        # same wavelengths as observed spectrum. However, if a fits name is 
        # provided in grid_name and that file exists, it is assumed the model 
        # grid in it is already resampled to match lbda_obs.

    if save and output_file is None:
        output_file = 'MCMC_results'

    
    # #########################################################################
    # Initialization of the MCMC variables                                    #
    # #########################################################################
    dim = len(labels)
    itermin = niteration_min
    limit = niteration_limit
    supp = niteration_supp
    maxgap = check_maxgap
    initial_state = np.array(initial_state)
    
    if itermin > limit:
        itermin = 0

    fraction = 0.3
    geom = 0
    lastcheck = 0
    konvergence = np.inf
    rhat_count = 0
    ac_count = 0
    chain = np.empty([nwalkers, 1, dim])
    ar_frac = np.empty([nwalkers, 1])
    nIterations = limit + supp
    rhat = np.zeros(dim)
    stop = np.inf
    
    # initialise ball of walkers
    if ini_ball == "uniform":
        pos = np.zeros([nwalkers, dim])
        for ii in range(dim):
            low_b = bounds[labels[ii]][0]
            up_b = (bounds[labels[ii]][1])
            pos[:,ii] = np.random.uniform(low_b + 0.01*(up_b-low_b), 
                                          up_b, nwalkers)
    elif isinstance(ini_ball,float):
        pos = initial_state + np.random.normal(0, ini_ball, (nwalkers, dim))
    else:
        raise ValueError("ini_ball must be string or float")
    
    sampler = emcee.EnsembleSampler(nwalkers, dim, spec_lnprob, a,
                                    args=([labels, bounds, grid_param_list, 
                                           lbda_obs, spec_obs, err_obs, dist,
                                           model_grid, model_reader, em_lines,
                                           em_grid, dlbda_obs, instru_corr, 
                                           instru_fwhm,instru_idx,filter_reader, 
                                           AV_bef_bb, units_obs, units_mod, 
                                           interp_order, priors, physical]),
                                    threads=nproc)
                                    
    start = datetime.datetime.now()

    # #########################################################################
    # Affine Invariant MCMC run
    # #########################################################################
    if verbosity > 1:
        print('\nStart of the MCMC run ...')
        print('Step  |  Duration/step (sec)  |  Remaining Estimated Time (sec)')
    
    for k, res in enumerate(sampler.sample(pos, iterations=nIterations,
                                           storechain=True)):
        elapsed = (datetime.datetime.now()-start).total_seconds()
        if verbosity > 1:
            if k == 0:
                q = 0.5
            else:
                q = 1
            print('{}\t\t{:.5f}\t\t\t{:.5f}'.format(k, elapsed * q,
                                                    elapsed * (limit-k-1) * q))
            
        start = datetime.datetime.now()

        # ---------------------------------------------------------------------
        # Store the state manually in order to handle with dynamical sized chain
        # ---------------------------------------------------------------------
        # Check if the size of the chain is long enough.
        s = chain.shape[1]
        if k+1 > s:     # if not, one doubles the chain length
            empty = np.zeros([nwalkers, 2*s, dim])
            chain = np.concatenate((chain, empty), axis=1)
            empty = np.zeros([nwalkers, 2*s])
            ar_frac = np.concatenate((ar_frac, empty), axis=1)
        # Store the state of the chain
        chain[:, k] = res[0]
        ar_frac[:,k] = sampler.acceptance_fraction

        # ---------------------------------------------------------------------
        # If k meets the criterion, one tests the non-convergence.
        # ---------------------------------------------------------------------
        criterion = int(np.amin([np.ceil(itermin*(1+fraction)**geom),
                        lastcheck+np.floor(maxgap)]))
        if k == criterion:
            
            geom += 1
            lastcheck = k
            if display:
                spec_show_walk_plot(chain)
                
            # We only test the rhat if we have reached the min # of steps
            if (k+1) >= itermin and konvergence == np.inf:
                if verbosity > 1:
                    print('\n{} convergence test in progress...'.format(conv_test))
                if conv_test == 'gb':
                    thr0 = int(np.floor(burnin*k))
                    thr1 = int(np.floor((1-burnin)*k*0.25))
    
                    # We calculate the rhat for each model parameter.
                    for j in range(dim):
                        part1 = chain[:, thr0:thr0 + thr1, j].reshape(-1)
                        part2 = chain[:, thr0 + 3 * thr1:thr0 + 4 * thr1, j
                                     ].reshape(-1)
                        series = np.vstack((part1, part2))
                        rhat[j] = gelman_rubin(series)
                    if verbosity > 0:
                        print('   r_hat = {}'.format(rhat))
                        cond = rhat <= rhat_threshold
                        print('   r_hat <= threshold = {} \n'.format(cond))
                    # We test the rhat.
                    if (rhat <= rhat_threshold).all():
                        rhat_count += 1
                        if rhat_count < rhat_count_threshold:
                            if verbosity > 0:
                                msg = "Gelman-Rubin test OK {}/{}"
                                print(msg.format(rhat_count, 
                                                 rhat_count_threshold))
                        elif rhat_count >= rhat_count_threshold:
                            if verbosity > 0:
                                print('... ==> convergence reached')
                            konvergence = k
                            stop = konvergence + supp
                    else:
                        rhat_count = 0
                elif conv_test == 'ac':
                    # We calculate the auto-corr test for each model parameter.
                    for j in range(dim):
                        rhat[j] = autocorr_test(chain[:,:k,j])
                    thr = 1./ac_c
                    if verbosity > 0:
                        print('Auto-corr tau/N = {}'.format(rhat))
                        print('tau/N <= {} = {} \n'.format(thr, rhat<thr))
                    if (rhat <= thr).all():
                        ac_count+=1
                        if verbosity > 0:
                            msg = "Auto-correlation test passed for all params!"
                            msg+= "{}/{}".format(ac_count,ac_count_thr)
                            print(msg)
                        if ac_count >= ac_count_thr:
                            msg='\n ... ==> convergence reached'
                            print(msg)
                            stop = k
                    else:
                        ac_count = 0
                else:
                    raise ValueError('conv_test value not recognized')
                
                                
                if save and verbosity == 3:
                    ac_time = []
                    for j in range(dim):
                        ac_time.append(k*autocorr_test(chain[:,:k,j]))
                    fname = '{d}/{f}_temp_k{k}'.format(d=output_dir,
                                                       f=output_file, k=k)
                    data = {'chain': sampler.chain,
                            'lnprob': sampler.lnprobability,
                            'ac_time': ac_time,
                             'AR': ar_frac}
                    with open(fname, 'wb') as fileSave:
                        pickle.dump(data, fileSave)

        # We have reached convergence
        if k+1 >= stop:
            if verbosity > 0:
                print('We break the loop because we have reached convergence')
            break
        # We have reached the maximum number of steps for our Markov chain.
        if k+1 >= nIterations-1:
            # break to avoid a bug related to font type
            break
        
    isamples, ln_proba, ar = spec_chain_zero_truncated(chain, 
                                                       sampler.lnprobability,
                                                       ar_frac)   
    # update units in the chain if needed
    if len(em_grid)>0:
        for key, val in em_lines.items():
            idx_line = labels.index(key)
            if val[1] == 'L':
                R_si = isamples[:,:,idx_R]*con.R_jup.value
                isamples[:,:,idx_line] *= 4*np.pi*R_si**2
            elif val[1] == 'LogL':
                R_si = isamples[:,:,idx_R]*con.R_jup.value
                conv_fac = 4*np.pi*R_si**2/con.L_sun.value
                isamples[:,:,idx_line] = np.log10(isamples[:,:,idx_line]*conv_fac)
      
    if k == nIterations-1:
        if verbosity > 0:
            print("We have reached the limit # of steps without convergence")
            print("You may have to increase niteration_limit")
    # #########################################################################
    # Construction of the independent samples
    # #########################################################################
 

    if save:
        frame = inspect.currentframe()
        args, _, _, values = inspect.getargvalues(frame)
        input_parameters = {j: values[j] for j in args[1:]}
        ac_time = []
        for j in range(dim):
            ac_time.append(isamples.shape[1]*autocorr_test(isamples[:,:,j]))
        output = {'isamples': isamples,
                  #'chain': chain,
                  'input_parameters': input_parameters,
                  'AR': ar,
                  'ac_time': ac_time,
                  'lnprobability': ln_proba}
                  
        with open(output_dir+output_file, 'wb') as fileSave:
            pickle.dump(output, fileSave)
        
        msg = "\nThe file MCMC_results has been stored in the folder {}"
        print(msg.format(output_dir))

    if verbosity > 0:
        timing(start_time)
                                    
    return isamples, ln_proba

                                    
def spec_chain_zero_truncated(chain, ln_proba=None, ar=None):
    """
    Return the Markov chain with the dimension: walkers x steps* x parameters,
    where steps* is the last step before having 0 (not yet constructed chain).
    
    Parameters
    ----------
    chain: numpy.array
        The MCMC chain.
    ln_proba: numpy.array, opt
        Corresponding ln-probabilities.
     
    Returns
    -------
    out: numpy.array
        The truncated MCMC chain, that is to say, the chain which only contains
        relevant information.
    out_ln_proba: numpy.array
        If ln_proba is provided as input, out_ln_proba contains the 
        zero-truncated ln-proba (i.e. matching shape with non-zero samples)
    """
    try:
        idxzero = np.where(chain[0, :, 0] == 0.0)[0][0]
    except:
        idxzero = chain.shape[1]
    res = [chain[:, 0:idxzero, :]]
        
    if ln_proba is not None:
        res.append(ln_proba[:,0:idxzero])
    if ar is not None:
        res.append(ar[:,0:idxzero])
    if len(res) == 1:
        return res[0]
    else:
        return tuple(res)
 
   
def spec_show_walk_plot(chain, labels, save=False, output_dir='', ntrunc=100,
                        **kwargs):
    """
    Display/save a figure showing the path of each walker during the MCMC run.
    
    Parameters
    ----------
    chain: numpy.array
        The Markov chain. The shape of chain must be nwalkers x length x dim.
        If a part of the chain is filled with zero values, the method will
        discard these steps.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - next the planet photometric radius 'R',
        - (optionally) the optical extinction 'Av'.
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution
    save: boolean, default: False
        If True, a pdf file is created.
    ntrunc: int, opt
        max number of walkers plotted. If too many walkers the plot will become
        too voluminous and too crowded. Plot will be truncated to only ntrunc 
        first walkers
    kwargs:
        Additional attributes are passed to the matplotlib plot method.
                                                        
    Returns
    -------
    Display the figure or create a pdf file named walk_plot.pdf in the working
    directory.
    
    """
    npar = chain.shape[-1]
    if len(labels) != npar:
        raise ValueError("Labels length should match chain last dimension")
    temp = np.where(chain[0, :, 0] == 0.0)[0]
    if len(temp) != 0:
        chain = chain[:, :temp[0], :]


    color = kwargs.pop('color', 'k')
    alpha = kwargs.pop('alpha', 0.1)
    filename = kwargs.pop('filename','walk_plot.pdf')
    #labels = kwargs.pop('labels', ["$r$", r"$\theta$", "$f$"])
    fig, axes = plt.subplots(npar, 1, sharex=True,
                             figsize=kwargs.pop('figsize', (8, int(2*npar))))
    if npar>1:
        axes[-1].set_xlabel(kwargs.pop('xlabel', 'step number'))
        axes[-1].set_xlim(kwargs.pop('xlim', [0, chain.shape[1]]))
        for j in range(npar):
            axes[j].plot(chain[:ntrunc, :, j].T, color=color, alpha=alpha, **kwargs)
            axes[j].yaxis.set_major_locator(MaxNLocator(5))
            axes[j].set_ylabel(labels[j])
    else:
        axes.set_xlabel(kwargs.pop('xlabel', 'step number'))
        axes.set_xlim(kwargs.pop('xlim', [0, chain.shape[1]]))            
        for j in range(npar):
            axes.plot(chain[:ntrunc, :, j].T, color=color, alpha=alpha, **kwargs)
            axes.yaxis.set_major_locator(MaxNLocator(5))
            axes.set_ylabel(labels[j])
    fig.tight_layout(h_pad=0)
    if save:
        plt.savefig(output_dir+filename)
        plt.close(fig)
    else:
        plt.show()


def spec_show_corner_plot(chain, burnin=0.5, save=False, output_dir='', 
                          mcmc_res=None, units=None, ndig=None, 
                          labels_plot=None, plot_name='corner_plot.pdf', **kwargs):
    """
    Display/save a figure showing the corner plot (pdfs + correlation plots).
    
    Parameters
    ----------
    chain: numpy.array
        The Markov chain. The shape of chain must be nwalkers x length x dim.
        If a part of the chain is filled with zero values, the method will
        discard these steps.
    burnin: float, default: 0
        The fraction of a walker we want to discard.
    save: boolean, default: False
        If True, a pdf file is created.
    mcmc_res: numpy array OR tuple of 2 dictionaries/np.array, opt
        Values to be printed on top of each 1d posterior distribution   
        - if numpy array: 
            - npar x 3 dimensions (where npar is the number of parameters), 
            containing the most likely value of each parameter and the lower 
            and upper uncertainties at the desired quantiles, resp. 
            - npar x 2 dimensions: same as above but with a single value of  
            uncertainty. E.g. output of spec_confidence() for a gaussian fit
        - if tuple of 2 dictionaries: output of spec_confidence without 
                                      gaussian fit
        - if tuple of 2 np.array: output of spec_confidence() with gaussian fit
    units: tuple, opt
        Tuple of strings containing units for each parameter. If provided,
        mcmc_res will be printed on top of each 1d posterior distribution along 
        with these units.
    ndig: tuple, opt
        Number of digits precision for each printed parameter
    labels_plot: tuple, opt
        Labels corresponding to parameter names, used for the plot. If None,
        will use "labels" passed in kwargs.
    kwargs:
        Additional attributes passed to the corner.corner() method.
        (e.g. 'labels', 'show_title')
                    
    Returns
    -------
    Display the figure or create a pdf file named walk_plot.pdf in the working
    directory.
        
    Raises
    ------
    ImportError
    
    """
    
    npar = chain.shape[-1]
    if "labels" in kwargs.keys():
        labels = kwargs.pop('labels')
        if labels_plot is None:
            labels_plot = labels
    else:
        labels = None
        
    # allows for different title/units on top of 1D posterior distributions
    labels_tit = kwargs.pop('labels_tit', labels_plot)
    labels_tit_unit = kwargs.pop('labels_tit_unit', units)
    
    try:
        temp = np.where(chain[0, :, 0] == 0.0)[0]
        if len(temp) != 0:
            chain = chain[:, :temp[0], :]
        length = chain.shape[1]
        indburn = int(np.floor(burnin*(length-1)))
        chain = chain[:, indburn:length, :].reshape((-1, npar))
    except IndexError:
        pass

    if chain.shape[0] == 0:
        raise ValueError("It seems the chain is empty. Have you run the MCMC?")
    else:
        fig = corner.corner(chain, labels=labels_plot, **kwargs)

        
    if mcmc_res is not None and labels is not None:
        title_kwargs = kwargs.pop('title_kwargs', None)
        if isinstance(mcmc_res,tuple):
            if len(mcmc_res) != 2:
                msg = "mcmc_res should have 2 elements"
                raise TypeError(msg)
            if len(mcmc_res[0]) != npar or len(mcmc_res[1]) != npar:
                msg = "dict should have as many entries as there are params"
                raise TypeError(msg)
            if labels is None:
                msg = "labels should be provided"
                raise TypeError(msg)                
        elif isinstance(mcmc_res,np.ndarray):
            if mcmc_res.shape[0] != npar:
                msg = "mcmc_res first dim should be equal to number of params"
                raise TypeError(msg)  
        else:
            msg = "type of mcmc_res not recognised"
            raise TypeError(msg)
        # Extract the axes
        axes = np.array(fig.axes).reshape((npar, npar))
                
        # Loop over the diagonal
        for i in range(npar):
            ax = axes[i, i]
            title = None
            if isinstance(mcmc_res,tuple):
                if isinstance(mcmc_res[0],dict):
                    q_50 = mcmc_res[0][labels[i]]
                    q_m = mcmc_res[1][labels[i]][0]
                    q_p = mcmc_res[1][labels[i]][1]
                else:
                    q_50 = mcmc_res[0][i][0]
                    q_m = mcmc_res[1][i][0]
                    q_p = q_m
            else:
                q_50 = mcmc_res[i,0]
                q_m = mcmc_res[i,1]
                q_p = mcmc_res[i,2]

            # Format the quantile display.
            try:
                fmt = "{{:.{0}f}}".format(ndig[i]).format
            except:
                fmt = "{{:.2f}}".format
            title = r"${{{0}}}_{{{1}}}^{{+{2}}}$"
            title = title.format(fmt(q_50), fmt(q_m), fmt(q_p))

            # Add in the column name if it's given.
            try:
                title = "{0} = {1} {2}".format(labels_tit[i], title, 
                                               labels_tit_unit[i])
            except:
                title = "{0} = {1}".format(labels_tit[i], title)
                
            ax.set_title(title, **title_kwargs)
    
    if save:
        plt.savefig(output_dir+plot_name)
        plt.close(fig)
    else:
        plt.show()


def spec_confidence(isamples, labels, cfd=68.27, bins=100, gaussian_fit=False, 
                    weights=None, verbose=True, save=False, output_dir='', 
                    bounds=None, priors=None, **kwargs):
    """
    Determine the highly probable value for each model parameter, as well as
    the 1-sigma confidence interval.
    
    Parameters
    ----------
    isamples: numpy.array
        The independent samples for each model parameter.
    labels: Tuple of strings
        Tuple of labels in the same order as initial_state, that is:
        - first all parameters related to loaded models (e.g. 'Teff', 'logg')
        - next the planet photometric radius 'R',
        - (optionally) the optical extinction 'Av'.
        - (optionally) 'Tbb1', 'Rbb1', 'Tbb2', 'Rbb2', etc. for each extra bb
        contribution
    cfd: float, optional
        The confidence level given in percentage.
    bins: int, optional
        The number of bins used to sample the posterior distributions.
    gaussian_fit: boolean, optional
        If True, a gaussian fit is performed in order to determine (\mu,\sigma)
    weights : (n, ) numpy ndarray or None, optional
        An array of weights for each sample.
    verbose: boolean, optional
        Display information in the shell.
    save: boolean, optional
        If "True", a txt file with the results is saved in the output
        repository.
    bounds: dictionary, opt
        Only used if a text file is saved summarizing results+bounds+priors.
        Should be the same bounds as provided to the MCMC.
    priors: dictionary, opt
        Only used if a text file is saved summarizing results+bounds+priors.
        Should be the same priors as provided to the MCMC.
    kwargs: optional
        Additional attributes are passed to the matplotlib hist() method.
        
    Returns
    -------
    out: tuple
        A 2 elements tuple with the highly probable solution and the confidence
        interval.
        
    """
    title = kwargs.pop('title', None)
        
    output_file = kwargs.pop('filename', 'confidence.txt')
    
    if gaussian_fit:
        output_pdf = kwargs.pop('pdfname','confi_hist_spec_params_gaussfit.pdf')
    else:
        output_pdf = kwargs.pop('pdfname','confi_hist_spec_params.pdf')
        
    try:
        l = isamples.shape[1]
    except:
        l = 1
     
    confidenceInterval = {}
    val_max = {}
    pKey = labels #['r', 'theta', 'f']
    
    if l != len(pKey):
        raise ValueError("Labels length should match chain last dimension")
    
    if cfd == 100:
        cfd = 99.9
        
    #########################################
    ##  Determine the confidence interval  ##
    #########################################
    if gaussian_fit:
        mu = np.zeros(l)
        sigma = np.zeros_like(mu)
    
    if gaussian_fit:
        fig, ax = plt.subplots(2, l, figsize=(int(4*l),8))
    else:
        fig, ax = plt.subplots(1, l, figsize=(int(4*l),4))
    
    for j in range(l):
        
        if gaussian_fit:
            n, bin_vertices, _ = ax[0][j].hist(isamples[:,j], bins=bins,
                                               weights=weights, histtype='step',
                                               edgecolor='gray')
        else:
            n, bin_vertices, _ = ax[j].hist(isamples[:,j], bins=bins,
                                            weights=weights, histtype='step',
                                            edgecolor='gray')
        bins_width = np.mean(np.diff(bin_vertices))
        surface_total = np.sum(np.ones_like(n)*bins_width * n)
        n_arg_sort = np.argsort(n)[::-1]
        
        test = 0
        pourcentage = 0
        for k, jj in enumerate(n_arg_sort):
            test = test + bins_width*n[int(jj)]
            pourcentage = test/surface_total*100
            if pourcentage > cfd:
                if verbose:
                    msg = 'percentage for {}: {}%'
                    print(msg.format(pKey[j], pourcentage))
                break
        n_arg_min = int(n_arg_sort[:k].min())
        n_arg_max = int(n_arg_sort[:k+1].max())
        
        if n_arg_min == 0:
            n_arg_min += 1
        if n_arg_max == bins:
            n_arg_max -= 1
        
        val_max[pKey[j]] = bin_vertices[int(n_arg_sort[0])]+bins_width/2.
        confidenceInterval[pKey[j]] = np.array([bin_vertices[n_arg_min-1],
                                               bin_vertices[n_arg_max+1]]
                                               - val_max[pKey[j]])
                        
        arg = (isamples[:, j] >= bin_vertices[n_arg_min - 1]) * \
              (isamples[:, j] <= bin_vertices[n_arg_max + 1])
        if gaussian_fit:
            ax[0][j].hist(isamples[arg,j], bins=bin_vertices,
                          facecolor='gray', edgecolor='darkgray',
                          histtype='stepfilled', alpha=0.5)
            ax[0][j].vlines(val_max[pKey[j]], 0, n[int(n_arg_sort[0])],
                            linestyles='dashed', color='red')
            ax[0][j].set_xlabel(pKey[j])
            if j == 0:
                ax[0][j].set_ylabel('Counts')

            mu[j], sigma[j] = norm.fit(isamples[:, j])
            n_fit, bins_fit = np.histogram(isamples[:, j], bins, normed=1,
                                           weights=weights)
            ax[1][j].hist(isamples[:, j], bins, density=1, weights=weights,
                          facecolor='gray', edgecolor='darkgray',
                          histtype='step')
            y = norm.pdf(bins_fit, mu[j], sigma[j])
            ax[1][j].plot(bins_fit, y, 'r--', linewidth=2, alpha=0.7)

            ax[1][j].set_xlabel(pKey[j])
            if j == 0:
                ax[1][j].set_ylabel('Counts')

            if title is not None:
                msg = r"{}   $\mu$ = {:.4f}, $\sigma$ = {:.4f}"
                ax[1][j].set_title(msg.format(title, mu[j], sigma[j]),
                                   fontsize=10)

        else:
            ax[j].hist(isamples[arg,j],bins=bin_vertices, facecolor='gray',
                       edgecolor='darkgray', histtype='stepfilled',
                       alpha=0.5)
            ax[j].vlines(val_max[pKey[j]], 0, n[int(n_arg_sort[0])],
                         linestyles='dashed', color='red')
            ax[j].set_xlabel(pKey[j])
            if j == 0:
                ax[j].set_ylabel('Counts')

            if title is not None:
                msg = r"{} - {:.3f} {:.3f} +{:.3f}"
                ax[1].set_title(msg.format(title, val_max[pKey[j]], 
                                           confidenceInterval[pKey[j]][0], 
                                           confidenceInterval[pKey[j]][1]),
                                fontsize=10)

        plt.tight_layout(w_pad=0.1)

    if save:
        plt.savefig(output_dir+output_pdf)


    if verbose:
        for j in range(l):
            print("******* Results for {} ***** ".format(pKey[j]))
            print('\n\nConfidence intervals:')
            print('{}: {} [{},{}]'.format(pKey[j], val_max[pKey[j]],
                                          confidenceInterval[pKey[j]][0],
                                          confidenceInterval[pKey[j]][1]))
            if gaussian_fit:
                print('Gaussian fit results:')
                print('{}: {} +-{}'.format(pKey[j], mu[j], sigma[j]))

    ##############################################
    ##  Write inference results in a text file  ##
    ##############################################
    if save:
        with open(output_dir+output_file, "w") as f:
            f.write('######################################\n')
            f.write('####   MCMC results (confidence)   ###\n')
            f.write('######################################\n')
            f.write(' \n')
            f.write('Results of the MCMC fit\n')
            f.write('----------------------- \n')
            f.write(' \n')
            if bounds is not None:
                f.write('Bounds\n')
                f.write('------')
                for j in range(l):
                    f.write('\n{}: [{},{}]'.format(pKey[j], bounds[pKey[j]][0],
                                                   bounds[pKey[j]][1]))
                f.write(' \n')
                f.write(' \n')
            f.write('Priors\n')
            f.write('------')
            if priors is not None:
                for key, prior in priors.items():
                    f.write('\n{}: {}+-{}'.format(key, prior[0], prior[1]))
            else:
                f.write('\n None')
            f.write(' \n')
            f.write(' \n')
            f.write('>> Spectral parameters for a ')
            f.write('{} % confidence interval:\n'.format(cfd))
            
            for j in range(l):
                f.write('{}: {} [{},{}]'.format(pKey[j], val_max[pKey[j]],
                                              confidenceInterval[pKey[j]][0],
                                              confidenceInterval[pKey[j]][1]))
                if gaussian_fit:
                    f.write('Gaussian fit results:')
                    f.write('{}: {} +-{}'.format(pKey[j], mu[j], sigma[j]))

    if gaussian_fit:
        return mu, sigma
    else:
        return val_max, confidenceInterval



#! /usr/bin/env python

"""
Module with HCIDataset and HCIFrame classes.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['Dataset',
           'Frame']

import numpy as np
import copy
import hciplot as hp
from .fits import open_fits
from .preproc import (frame_crop, frame_px_resampling, frame_rotate,
                      frame_shift, frame_center_satspots, frame_center_radon)
from .preproc import (cube_collapse, cube_crop_frames, cube_derotate,
                      cube_drop_frames, cube_detect_badfr_correlation,
                      cube_detect_badfr_pxstats, cube_detect_badfr_ellipticity,
                      cube_px_resampling, cube_subsample, cube_recenter_2dfit,
                      cube_recenter_satspots, cube_recenter_radon,
                      cube_recenter_dft_upsampling, cube_recenter_via_speckles)
from .var import (frame_filter_lowpass, frame_filter_highpass, frame_center,
                  cube_filter_highpass, cube_filter_lowpass, mask_circle)
from .stats import (frame_basic_stats, frame_histo_stats,
                    frame_average_radprofile, cube_basic_stats, cube_distance)
from .metrics import (frame_report, cube_inject_companions,
                      generate_cube_copies_with_injections, snr,
                      snrmap, detection, normalize_psf)

from .conf.utils_conf import check_array, Saveable, print_precision
from .conf.mem import check_enough_memory


class Frame(object):
    """ High-contrast imaging frame (2d array).

    Parameters
    ----------
    data : numpy ndarray
        2d array.
    hdu : int, optional
        If ``cube`` is a String, ``hdu`` indicates the HDU from the FITS file.
        By default the first HDU is used.
    fwhm : float, optional
        The FWHM associated with this dataset (instrument dependent). Required
        for several methods (operations on the cube).
    """
    def __init__(self, data, hdu=0, fwhm=None):
        """ HCIFrame object initialization. """
        if isinstance(data, str):
            self.data = open_fits(data, hdu, verbose=False)
        else:
            self.data = data

        check_array(self.data, dim=2, msg='Image.data')
        print('Frame shape: {}'.format(self.data.shape))

        self.fwhm = fwhm
        if self.fwhm is not None:
            print('FWHM: {}'.format(self.fwhm))

    def crop(self, size, xy=None, force=False):
        """ Cropping the frame.

        Parameters
        ----------
        size : int, odd
            Size of the subframe.
        cenxy : tuple, optional
            Coordinates of the center of the subframe.
        force : bool, optional
            Size and the size of the 2d array must be both even or odd. With
            ``force`` set to True this condition can be avoided.

        """
        self.data = frame_crop(self.data, size, xy, force, verbose=True)

    def detect_blobs(self, psf, bkg_sigma=1, method='lpeaks',
                     matched_filter=False, mask=True, snr_thresh=5, plot=True,
                     debug=False, verbose=False, save_plot=None,
                     plot_title=None, angscale=False):
        """ Detecting blobs on the 2d array.
        """
        self.detection_results = detection(self.data, psf, bkg_sigma, method,
                                           matched_filter, mask, snr_thresh,
                                           plot, debug, True, verbose,
                                           save_plot, plot_title, angscale)

    def filter(self, method, mode, median_size=5, kernel_size=5, fwhm_size=5,
               btw_cutoff=0.2, btw_order=2, gauss_mode='conv'):
        """ High/low pass filtering the frames of the image.

        Parameters
        ----------
        method : {'lp', 'hp'}

        mode : {'median', 'gauss'}
        {'laplacian', 'laplacian-conv', 'median-subt', 'gauss-subt', 'fourier-butter'}
        """
        if method == 'hp':
            self.data = frame_filter_highpass(self.data, mode, median_size,
                                              kernel_size, fwhm_size,
                                              btw_cutoff, btw_order)
        elif method == 'lp':
            self.data = frame_filter_lowpass(self.data, mode, median_size,
                                             fwhm_size, gauss_mode)
        else:
            raise ValueError('Filtering mode not recognized')
        print('Image successfully filtered')

    def get_center(self, verbose=True):
        """ Getting the coordinates of the center of the image.

        Parameters
        ----------
        verbose : bool optional
            If True the center coordinates are printed out.
        """
        return frame_center(self.data, verbose)

    def plot(self, **kwargs):
        """ Plotting the 2d array.

        Parameters
        ----------
        **kwargs : dict, optional
            Parameters passed to the function ``plot_frames`` of the package
            ``HCIplot``.
        """
        hp.plot_frames(self.data, **kwargs)

    def radial_profile(self, sep=1):
        """ Calculates the average radial profile of an image.

        Parameters
        ----------
        sep : int, optional
            The average radial profile is recorded every ``sep`` pixels.
        """
        radpro = frame_average_radprofile(self.data, sep=sep, plot=True)
        return radpro

    def recenter(self, method='satspots', xy=None, subi_size=19, sigfactor=6,
                 imlib='opencv', interpolation='lanczos4', debug=False,
                 verbose=True):
        """ Recentering the frame using the satellite spots or a radon
        transform.

        Parameters
        ----------
        method : {'satspots', 'radon'}, str optional
            Method for recentering the frame.
        xy : tuple, optional
            Tuple with coordinates X,Y of the satellite spots. When the spots
            are in an X configuration, the order is the following: top-left,
            top-right, bottom-left and bottom-right. When the spots are in an
            + (cross-like) configuration, the order is the following: top,
            right, left, bottom.
        """
        if method == 'satspots':
            if xy is None:
                raise ValueError('`xy` must be a tuple of 4 tuples')
            self.data, _, _ = frame_center_satspots(self.data, xy, subi_size,
                                                    sigfactor, True, imlib,
                                                    interpolation, debug,
                                                    verbose)
        elif method == 'radon':
            pass
            # self.data = frame_center_radon()
        else:
            raise ValueError('Recentering method not recognized')

    def rescale(self, scale, imlib='ndimage', interpolation='bicubic',
                verbose=True):
        """ Resampling the image (upscaling or downscaling).

        Parameters
        ----------
        scale : int, float or tuple
            Scale factor for upsampling or downsampling the frames in the cube.
            If a tuple it corresponds to the scale along x and y.
        imlib : {'ndimage', 'opencv'}, str optional
            Library used for image transformations. ndimage is the default.
        interpolation : str, optional
            For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation
            is the fastest and the 'biquintic' the slowest. The 'nearneig' is
            the worst option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate.
        verbose : bool, optional
            Whether to print out additional info such as the new image shape.
        """
        self.data = frame_px_resampling(self.data, scale, imlib, interpolation,
                                        verbose)

    def rotate(self, angle, imlib='opencv', interpolation='lanczos4', cxy=None):
        """ Rotating the image by a given ``angle``.

        Parameters
        ----------
        imlib : {'opencv', 'skimage'}, str optional
            Library used for image transformations. Opencv is faster than
            ndimage or skimage.
        interpolation : str, optional
            For 'skimage' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation
            is the fastest and the 'biquintic' the slowest. The 'nearneig' is
            the poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        cxy : tuple of int, optional
            Coordinates X,Y  of the point with respect to which the rotation
            will be performed. By default the rotation is done with respect to
            the center of the frames, as it is returned by the function
            vip_hci.var.frame_center.

        """
        self.data = frame_rotate(self.data, angle, imlib, interpolation, cxy)
        print('Image successfully rotated')

    def shift(self, shift_y, shift_x, imlib='opencv', interpolation='lanczos4'):
        """ Shifting the image.

        Parameters
        ----------
        shift_y, shift_x: float
            Shifts in x and y directions.
        imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
            Library or method used for performing the image shift.
            'ndimage-fourier', does a fourier shift operation and preserves
            better the pixel values (therefore the flux and photometry).
            Interpolation based shift ('opencv' and 'ndimage-interp') is faster
            than the fourier shift. 'opencv' is recommended when speed is
            critical.
        interpolation : {'bicubic', 'bilinear', 'nearneig'}, optional
            Only used in case of imlib is set to 'opencv' or 'ndimage-interp',
            where the images are shifted via interpolation.
            For 'ndimage-interp' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation is
            the fastest and the 'biquintic' the slowest. The 'nearneig' is the
            poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        """
        self.data = frame_shift(self.data, shift_y, shift_x, imlib,
                                interpolation)
        print('Image successfully shifted')

    def snr(self, source_xy, plot=False, verbose=True):
        """ Calculating the S/N for a test resolution element ``source_xy``.

        Parameters
        ----------
        source_xy : tuple of floats
            X and Y coordinates of the planet or test speckle.
        plot : bool, optional
            Plots the frame and the apertures considered for clarity.
        verbose : bool, optional
            Chooses whether to print some output or not.

        Returns
        -------
        snr_val : float
            Value of the S/N for ``source_xy``.
        """
        if self.fwhm is None:
            raise ValueError('FWHM has not been set')
        return snr(self.data, source_xy, self.fwhm, False, plot, verbose)

    def stats(self, region='circle', radius=5, xy=None, annulus_inner_radius=0,
              annulus_width=5, source_xy=None, verbose=True, plot=True):
        """ Calculating statistics on the image, both in the full-frame and in
        a region (circular aperture or annulus). Also, the S/N of the either
        ``source_xy`` or the max pixel is calculated.

        Parameters
        ----------
        region : {'circle', 'annulus'}, str optional
            Region in which basic statistics (mean, stddev, median and max) are
            calculated.
        radius : int, optional
            Radius of the circular aperture.
        xy : tuple of floats, optional
            Center of the circular aperture.
        annulus_inner_radius : int, optional
            Inner radius of the annular region.
        annulus_width : int, optional
            Width of the annular region.
        source_xy : tuple of floats, optional
            Coordinates for which the S/N information will be obtained. If None,
            the S/N is estimated for the pixel with the maximum value.
        verbose : bool, optional
            Whether to print out the values of the calculated statistics.
        plot : bool, optional
            Whether to plot the frame, histograms and region.
        """
        res_region = frame_basic_stats(self.data, region, radius, xy,
                                       annulus_inner_radius, annulus_width,
                                       plot, True)
        if verbose:
            if region == 'circle':
                msg = 'Stats in circular aperture of radius: {}pxs'
                print(msg.format(radius))
            elif region == 'annulus':
                msg = 'Stats in annulus. Inner_rad: {}pxs, width: {}pxs'
                print(msg.format(annulus_inner_radius, annulus_width))
            mean, std_dev, median, maxi = res_region
            msg = 'Mean: {:.3f}, Stddev: {:.3f}, Median: {:.3f}, Max: {:.3f}'
            print(msg.format(mean, std_dev, median, maxi))

        res_ff = frame_histo_stats(self.data, plot)
        if verbose:
            mean, median, std, maxim, minim = res_ff
            print('Stats in the whole frame:')
            msg = 'Mean: {:.3f}, Stddev: {:.3f}, Median: {:.3f}, Max: {:.3f}, '
            msg += 'Min: {:.3f}'
            print(msg.format(mean, std, median, maxim, minim))

        print('\nS/N info:')
        _ = frame_report(self.data, self.fwhm, source_xy, verbose)


class Dataset(Saveable):
    """ High-contrast imaging dataset class.

    Parameters
    ----------
    cube : str or numpy array
        3d or 4d high-contrast image sequence. If a string is provided, cube is
        interpreted as the path of the FITS file containing the sequence.
    hdu : int, optional
        If ``cube`` is a String, ``hdu`` indicates the HDU from the FITS file.
        By default the first HDU is used.
    angles : list or numpy array, optional
        The vector of parallactic angles.
    wavelengths : list or numpy array, optional
        The vector of wavelengths (to be used as scaling factors).
    fwhm : float, optional
        The FWHM associated with this dataset (instrument dependent). Required
        for several methods (operations on the cube).
    px_scale : float, optional
        The pixel scale associated with this dataset (instrument dependent).
    psf : numpy array, optional
        The PSF template associated with this dataset.
    psfn : numpy array, optional
        Normalized/cropped/centered version of the PSF template associated with
        this dataset.
    cuberef : str or numpy array
        3d or 4d high-contrast image sequence. To be used as a reference cube.
    """

    _saved_attributes = ["cube", "psf", "psfn", "angles", "fwhm", "wavelengths",
                         "px_scale", "cuberef", "injections_yx"]

    def __init__(self, cube, hdu=0, angles=None, wavelengths=None, fwhm=None,
                 px_scale=None, psf=None, psfn=None, cuberef=None):
        """ Initialization of the HCIDataset object.
        """
        # Loading the 3d/4d cube or image sequence
        if isinstance(cube, str):
            self.cube = open_fits(cube, hdu, verbose=False)
        elif isinstance(cube, np.ndarray):
            if not (cube.ndim == 3 or cube.ndim == 4):
                raise ValueError('`Cube` array has wrong dimensions')
            self.cube = cube
        else:
            raise TypeError('`Cube` has a wrong type')
        print('Cube array shape: {}'.format(self.cube.shape))
        if self.cube.ndim == 3:
            self.n, self.y, self.x = self.cube.shape
            self.w = 1
        elif self.cube.ndim == 4:
            self.w, self.n, self.y, self.x = self.cube.shape

        # Loading the reference cube
        if isinstance(cuberef, str):
            self.cuberef = open_fits(cuberef, hdu, verbose=False)
        elif isinstance(cuberef, np.ndarray):
            msg = '`Cuberef` array has wrong dimensions'
            if not cuberef.ndim == 3:
                raise ValueError(msg)
            if not cuberef.shape[1] == self.y:
                raise ValueError(msg)
            self.cuberef = cuberef
        elif isinstance(cuberef, Dataset):
            msg = '`Cuberef` array has wrong dimensions'
            if not cuberef.cube.ndim == 3:
                raise ValueError(msg)
            if not cuberef.cube.shape[1] == self.y:
                raise ValueError(msg)
            self.cuberef = cuberef.cube
        else:
            self.cuberef = None
        if self.cuberef is not None:
            print('Cuberef array shape: {}'.format(self.cuberef.shape))

        # Loading the angles (ADI)
        if isinstance(angles, str):
            self.angles = open_fits(angles, verbose=False)
        else:
            self.angles = angles
        if self.angles is not None:
            print('Angles array shape: {}'.format(self.angles.shape))
            # Checking the shape of the angles vector
            check_array(self.angles, dim=1, msg='Parallactic angles vector')
            if not self.angles.shape[0] == self.n:
                raise ValueError('Parallactic angles vector has a wrong shape')

        # Loading the scaling factors (mSDI)
        if isinstance(wavelengths, str):
            self.wavelengths = open_fits(wavelengths, verbose=False)
        else:
            self.wavelengths = wavelengths
        if self.wavelengths is not None:
            print('Wavelengths array shape: {}'.format(self.wavelengths.shape))
            # Checking the shape of the scaling vector
            check_array(self.wavelengths, dim=1, msg='Wavelengths vector')
            if not self.wavelengths.shape[0] == self.w:
                raise ValueError('Wavelengths vector has a wrong shape')

        # Loading the PSF
        if isinstance(psf, str):
            self.psf = open_fits(psf, verbose=False)
        else:
            self.psf = psf
        if self.psf is not None:
            print('PSF array shape: {}'.format(self.psf.shape))
            # Checking the shape of the PSF array
            if not self.psf.ndim == self.cube.ndim - 1:
                msg = 'PSF array has a wrong shape. Must have {} dimensions, '
                msg += 'got {} instead'
                raise ValueError(msg.format(self.cube.ndim - 1, self.psf.ndim))

        # Loading the normalized PSF
        if isinstance(psfn, str):
            self.psfn = open_fits(psfn, verbose=False)
        else:
            self.psfn = psfn
        if self.psfn is not None:
            print('Normalized PSF array shape: {}'.format(self.psfn.shape))
            # Checking the shape of the PSF array
            if not self.psfn.ndim == self.cube.ndim - 1:
                msg = 'Normalized PSF array has a wrong shape. Must have {} '
                msg += 'dimensions, got {} instead'
                raise ValueError(msg.format(self.cube.ndim - 1, self.psfn.ndim))

        self.fwhm = fwhm
        if self.fwhm is not None:
            if self.cube.ndim == 4:
                check_array(self.fwhm, dim=1, msg='FHWM')
            elif self.cube.ndim == 3:
                print('FWHM: {}'.format(self.fwhm))
        self.px_scale = px_scale
        if self.px_scale is not None:
            print('Pixel/plate scale: {}'.format(self.px_scale))

        self.injections_yx = None

    def collapse(self, mode='median', n=50):
        """ Collapsing the sequence into a 2d array.

        """
        frame = cube_collapse(self.cube, mode, n)
        print('Cube successfully collapsed')
        return Frame(frame)

    def crop_frames(self, size, xy=None, force=False):
        """ Cropping the frames of the sequence (3d or 4d cube).

        Parameters
        ----------
        size : int
            New size of the (square) frames.
        xy : tuple of ints
            X, Y coordinates of new frame center. If you are getting the
            coordinates from ds9 subtract 1, python has 0-based indexing.
        force : bool, optional
            ``Size`` and the original size of the frames must be both even or
            odd. With ``force`` set to True this condition can be avoided.
        """
        self.cube = cube_crop_frames(self.cube, size, xy, force, verbose=True)

    def derotate(self, imlib='opencv', interpolation='lanczos4', cxy=None,
                 nproc=1):
        """ Derotating the frames of the sequence according to the parallactic
        angles.

        Parameters
        ----------
        imlib : {'opencv', 'skimage'}, str optional
            Library used for image transformations. Opencv is faster than
            ndimage or skimage.
        interpolation : str, optional
            For 'skimage' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation
            is the fastest and the 'biquintic' the slowest. The 'nearneig' is
            the poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        cxy : tuple of int, optional
            Coordinates X,Y  of the point with respect to which the rotation
            will be performed. By default the rotation is done with respect to
            the center of the frames, as it is returned by the function
            vip_hci.var.frame_center.
        nproc : int, optional
            Whether to rotate the frames in the sequence in a multi-processing
            fashion. Only useful if the cube is significantly large (frame size
            and number of frames).
        """
        if self.angles is None:
            raise ValueError('Parallactic angles vector has not been set')

        self.cube = cube_derotate(self.cube, self.angles, imlib,
                                  interpolation, cxy, nproc)
        print('Cube successfully derotated')

    def drop_frames(self, n, m, verbose=True):
        """
        Slice the cube so that all frames between ``n``and ``m`` are kept.

        The indices ``n`` and ``m`` are included and 1-based.

        Examples
        --------
        For a cube which has 5 frames numbered ``1, 2, 3, 4, 5``, calling
        ``ds.drop_frames(2, 4)`` would result in the frames ``2, 3, 4`` to be
        kept, so the first and the last frame would be discarded.

        """
        res = cube_drop_frames(self.cube, n, m, self.angles, verbose=verbose)
        if self.angles:
            self.cube, self.angles = res
        else:
            self.cube = res

    def filter(self, method, mode, median_size=5, kernel_size=5, fwhm_size=5,
               btw_cutoff=0.2, btw_order=2, gauss_mode='conv', verbose=True):
        """ High/low pass filtering the frames of the cube.

        Parameters
        ----------
        method : {'lp', 'hp'}

        mode : {'median', 'gauss'}
        {'laplacian', 'laplacian-conv', 'median-subt', 'gauss-subt', 'fourier-butter'}
        """
        if method == 'hp':
            self.cube = cube_filter_highpass(self.cube, mode, median_size,
                                              kernel_size, fwhm_size,
                                              btw_cutoff, btw_order, verbose)
        elif method == 'lp':
            self.cube = cube_filter_lowpass(self.cube, mode, median_size,
                                             fwhm_size, gauss_mode, verbose)
        else:
            raise ValueError('Filtering mode not recognized')

    def frame_distances(self, frame, region='full', dist='sad',
                        inner_radius=None, width=None, plot=True):
        """ Calculating the frame distance/correlation with respect to a
        reference image.

        Parameters
        ----------
        frame : int or 2d array
            Reference frame in the cube or 2d array.
        region : {'full', 'annulus'}, string optional
            Whether to use the full frames or a centered annulus.
        dist : {'sad','euclidean','mse','pearson','spearman', 'ssim'}, str optional
            Which criterion to use.
        inner_radius : None or int, optional
            The inner radius when mode is 'annulus'.
        width : None or int, optional
            The width when mode is 'annulus'.
        plot : bool, optional
            Whether to plot the distances or not.

        """
        _ = cube_distance(self.cube, frame, region, dist, inner_radius, width,
                          plot)

    def frame_stats(self, region='circle', radius=5, xy=None,
                    annulus_inner_radius=0, annulus_width=5, wavelength=0,
                    plot=True):
        """ Calculating statistics on a ``region`` (circular aperture or
        annulus) of each image of the sequence.

        Parameters
        ----------
        region : {'circle', 'annulus'}, str optional
            Region in which basic statistics (mean, stddev, median and max) are
            calculated.
        radius : int, optional
            Radius of the circular aperture.
        xy : tuple of floats, optional
            Center of the circular aperture.
        annulus_inner_radius : int, optional
            Inner radius of the annular region.
        annulus_width : int, optional
            Width of the annular region.
        wavelength : int, optional
            Index of the wavelength to be analyzed in the case of a 4d cube.
        plot : bool, optional
            Whether to plot the frame, histograms and region.

        """
        if self.cube.ndim == 3:
            _ = cube_basic_stats(self.cube, region, radius, xy,
                                 annulus_inner_radius, annulus_width, plot,
                                 False)
        elif self.cube.ndim == 4:
            print('Stats for wavelength {}'.format(wavelength + 1))
            _ = cube_basic_stats(self.cube[wavelength], region, radius, xy,
                                 annulus_inner_radius, annulus_width, plot,
                                 False)

    def inject_companions(self, flux, rad_dists, n_branches=1, theta=0,
                          imlib='opencv', interpolation='lanczos4',
                          full_output=False, verbose=True):
        """ Injection of fake companions in 3d or 4d cubes.

        Parameters
        ----------
        flux : float or list
            Factor for controlling the brightness of the fake companions.
        rad_dists : float, list or array 1d
            Vector of radial distances of fake companions in pixels.
        n_branches : int, optional
            Number of azimutal branches.
        theta : float, optional
            Angle in degrees for rotating the position of the first branch that
            by default is located at zero degrees. Theta counts
            counterclockwise from the positive x axis.
        imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
            Library or method used for performing the image shift.
            'ndimage-fourier', does a fourier shift operation and preserves
            better the pixel values (therefore the flux and photometry).
            Interpolation based shift ('opencv' and 'ndimage-interp') is faster
            than the fourier shift. 'opencv' is recommended when speed is
            critical.
        interpolation : {'bicubic', 'bilinear', 'nearneig'}, optional
            Only used in case of imlib is set to 'opencv' or 'ndimage-interp',
            where the images are shifted via interpolation. For 'ndimage-interp'
            library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
            'biquartic', 'biquintic'. The 'nearneig' interpolation is the
            fastest and the 'biquintic' the slowest. The 'nearneig' is the
            poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        full_output : bool, optional
            Return the coordinates of the injected companions.
        verbose : bool, optional
            If True prints out additional information.

        Returns
        -------
        yx : list of tuple(y,x)
            [full_output=True] Pixel coordinates of the injections in the first
            frame (and first wavelength for 4D cubes). These are only the new
            injections - all injections (from multiple calls to this function)
            are stored in ``self.injections_yx``.

        """
        if self.angles is None:
            raise ValueError('The PA angles have not been set')
        if self.psfn is None:
            raise ValueError('The normalized PSF array cannot be found')
        if self.px_scale is None:
            raise ValueError('Pixel/plate scale has not been set')
        if self.cube.ndim == 4:
            if self.wavelengths is None:
                raise ValueError('The wavelengths vector has not been set')

        self.cube, yx = cube_inject_companions(
            self.cube, self.psfn, self.angles, flux, self.px_scale,
            rad_dists, n_branches, theta, imlib, interpolation,
            full_output=True, verbose=verbose
        )

        if self.injections_yx is None:
            self.injections_yx = []

        self.injections_yx += yx

        if verbose:
            print("Coordinates of the injections stored in self.injections_yx")

        if full_output:
            return yx

    def generate_copies_with_injections(self, n_copies, inrad=8, outrad=12,
                                        dist_flux=("uniform", 2, 500)):
        """
        Create copies of this dataset, containing different random injections.

        Parameters
        ----------
        n_copies : int
            This is the number of 'cube copies' returned.
        inrad,outrad : float
            Inner and outer radius of the injections. The actual injection
            position is chosen randomly.
        dist_flux : tuple('method', *params)
            Tuple describing the flux selection. Method can be a function, the
            ``*params`` are passed to it. Method can also be a string, for a
            pre-defined random function:

                ``("skewnormal", skew, mean, var)``
                    uses scipy.stats.skewnorm.rvs
                ``("uniform", low, high)``
                    uses np.random.uniform
                ``("normal", loc, scale)``
                    uses np.random.normal

        Yields
        -------
        fake_dataset : HCIDataset
            Copy of the original HCIDataset, with injected companions.

        """

        for data in generate_cube_copies_with_injections(
            self.cube, self.psf, self.angles, self.px_scale, n_copies=n_copies,
            inrad=inrad, outrad=outrad, dist_flux=dist_flux
        ):

            dsi = self.copy()
            dsi.cube = data["cube"]
            dsi.injections_yx = data["positions"]
            # data["dist"], data["theta"], data["flux"] are not used.

            yield dsi

    def get_nbytes(self):
        """
        Return the total number of bytes the HCIDataset consumes.
        """
        return sum(arr.nbytes for arr in [self.cube, self.cuberef, self.angles,
                                          self.wavelengths, self.psf, self.psfn]
                   if arr is not None)

    def copy(self, deep=True, check_mem=True):
        """
        Create an in-memory copy of this HCIDataset.

        This is especially useful for keeping a backup copy of the original
        dataset before modifying if (e.g. with injections).

        Parameters
        ----------
        deep : bool, optional
            By default, a deep copy is created. That means every (sub)attribute
            is copied in memory. While this requires more memory, one can safely
            modify the attributes without touching the original HCIDataset. When
            ``deep=False``, a shallow copy of the HCIDataset is returned
            instead. That means all attributes (e.g. ``self.cube``) point back
            to the original object's attributes. Pay attention when modifying
            such a shallow copy!
        check_mem : bool, optional
            [deep=True] If True, verifies that the system has enough memory to
            store the result.

        Returns
        -------
        new_dataset : Dataset
            (deep) copy of this HCIDataset.

        """
        if deep:
            if check_mem and not check_enough_memory(self.get_nbytes(), 1.5,
                                                     verbose=False):
                raise RuntimeError("copy would require more memory than "
                                   "available.")

            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def load_angles(self, angles, hdu=0):
        """ Loads the PA vector from a FITS file. It is possible to specify the
        HDU.

        Parameters
        ----------
        angles : str or 1d numpy ndarray
            List or vector with the parallactic angles.
        hdu : int, optional
            If ``angles`` is a String, ``hdu`` indicates the HDU from the FITS
            file. By default the first HDU is used.
        """
        if isinstance(angles, str):
            self.angles = open_fits(angles, hdu)
        elif isinstance(angles, (list, np.ndarray)):
            self.angles = angles
        else:
            msg = '`Angles` has a wrong type. Must be a list or 1d np.ndarray'
            raise ValueError(msg)

    def load_wavelengths(self, wavelengths, hdu=0):
        """ Loads the scaling factors vector from a FITS file. It is possible to
        specify the HDU.

        Parameters
        ----------
        wavelengths : str or 1d numpy ndarray
            List or vector with the wavelengths.
        hdu : int, optional
            If ``wavelengths`` is a String, ``hdu`` indicates the HDU from the
            FITS file. By default the first HDU is used.

        """
        if isinstance(wavelengths, str):
            self.wavelengths = open_fits(wavelengths, hdu)
        elif isinstance(wavelengths, (list, np.ndarray)):
            self.wavelengths = wavelengths
        else:
            msg = '`wavelengths` has a wrong type. Must be a list or np.ndarray'
            raise ValueError(msg)

    def mask_center(self, radius, fillwith=0, mode='in'):
        """ Masking the values inside/outside a centered circular aperture.

        Parameters
        ----------
        radius : int
            Radius of the circular aperture.
        fillwith : int, float or np.nan, optional
            Value to put instead of the masked out pixels.
        mode : {'in', 'out'}, optional
            When set to 'in' then the pixels inside the radius are set to
            ``fillwith``. When set to 'out' the pixels outside the circular
            mask are set to ``fillwith``.

        """
        self.cube = mask_circle(self.cube, radius, fillwith, mode)

    def normalize_psf(self, fit_fwhm=True, size=None, threshold=None,
                      mask_core=None, model='gauss', imlib='opencv',
                      interpolation='lanczos4', force_odd=True, verbose=True):
        """ Normalizes a PSF (2d or 3d array), to have the flux in a 1xFWHM
        aperture equal to one. It also allows to crop the array and center the
        PSF at the center of the frame(s).

        Parameters
        ----------
        fit_fwhm: bool, optional
            Whether to fit a ``model`` to estimate the FWHM instead of using the
            self.fwhm attribute.
        size : int or None, optional
            If int it will correspond to the size of the squared subimage to be
            cropped form the psf array.
        threshold : None of float, optional
            Sets to zero small values, trying to leave only the core of the PSF.
        mask_core : None of float, optional
            Sets the radius of a circular aperture for the core of the PSF,
            everything else will be set to zero.
        imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
            Library or method used for performing the image shift.
            'ndimage-fourier', does a fourier shift operation and preserves
            better the pixel values (therefore the flux and photometry).
            Interpolation based shift ('opencv' and 'ndimage-interp') is faster
            than the fourier shift. 'opencv' is recommended when speed is
            critical.
        interpolation : {'bicubic', 'bilinear', 'nearneig'}, optional
            Only used in case of imlib is set to 'opencv' or 'ndimage-interp',
            where the images are shifted via interpolation. For 'ndimage-interp'
            library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
            'biquartic', 'biquintic'. The 'nearneig' interpolation is the
            fastest and the 'biquintic' the slowest. The 'nearneig' is the
            poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        force_odd : str, optional
            If True the resulting array will have odd size (and the PSF will be
            placed at its center). If False, and the frame size is even, then
            the PSF will be put at the center of an even-sized frame.
        verbose : bool, optional
            If True intermediate results are printed out.
        """
        if not fit_fwhm and self.fwhm is None:
            raise ValueError('FWHM has not been set')
        if self.psf is None:
            raise ValueError('PSF array has not been loaded')

        if not fit_fwhm:
            fwhm = self.fwhm
        else:
            fwhm = 'fit'
        res = normalize_psf(self.psf, fwhm, size, threshold, mask_core, model,
                            imlib, interpolation, force_odd, True, verbose)
        self.psfn, self.aperture_flux, self.fwhm = res
        print('Normalized PSF array shape: {}'.format(self.psfn.shape))
        print('The attribute `psfn` contains the normalized PSF')
        print("`fwhm` attribute set to")
        print_precision(self.fwhm)

    def plot(self, **kwargs):
        """ Plotting the frames of a 3D or 4d cube.

        Parameters
        ----------
        **kwargs : dict, optional
            Parameters passed to the function ``plot_cubes`` of the package
            ``HCIplot``.
        """
        hp.plot_cubes(self.cube, **kwargs)

    def recenter(self, method='2dfit', xy=None, subi_size=5, model='gauss',
                 nproc=1, imlib='opencv', interpolation='lanczos4',
                 offset=None, negative=False, threshold=False,
                 save_shifts=False, cy_1=None, cx_1=None, upsample_factor=100,
                 alignment_iter=5, gamma=1, min_spat_freq=0.5, max_spat_freq=3,
                 recenter_median=False, sigfactor=6, cropsize=101, hsize=0.4,
                 step=0.01, mask_center=None, verbose=True, debug=False,
                 plot=True):
        """ Frame to frame recentering.

        Parameters
        ----------
        method : {'2dfit', 'dftups', 'dftupspeckles', 'satspots', 'radon'}, optional
            Recentering method.
        xy : tuple or ints or tuple of 4 tuples of ints, optional
            For the 2dfitting, ``xy`` are the coordinates of the center of the
            subimage (wrt the original frame). For the satellite spots method,
            it is a tuple with coordinates X,Y of the 4 satellite spots. When
            the spots are in an X configuration, the order is the following:
            top-left, top-right, bottom-left and bottom-right. When the spots
            are in an + (cross-like) configuration, the order is the following:
            top, right, left, bottom.
        subi_size : int, optional
            Size of the square subimage sides in pixels.
        model : str, optional
            [method=2dfit] Sets the type of fit to be used.
            'gauss' for a 2d Gaussian fit and 'moff' for a 2d Moffat fit.
        nproc : int or None, optional
            Number of processes (>1) for parallel computing. If 1 then it runs
            in serial. If None the number of processes will be set to
            (cpu_count()/2).
        imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
            Library or method used for performing the image shift.
            'ndimage-fourier', does a fourier shift operation and preserves
            better the pixel values (therefore the flux and photometry).
            Interpolation based shift ('opencv' and 'ndimage-interp') is faster
            than the fourier shift. 'opencv' is recommended when speed is
            critical.
        interpolation : {'bicubic', 'bilinear', 'nearneig'}, optional
            Only used in case of imlib is set to 'opencv' or 'ndimage-interp',
            where the images are shifted via interpolation.
            For 'ndimage-interp' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation is
            the fastest and the 'biquintic' the slowest. The 'nearneig' is the
            poorer option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate. 'lanczos4' is the default.
        offset : tuple of floats, optional
            [method=2dfit] If None the region of the frames
            used for the 2d Gaussian/Moffat fit is shifted to the center of the
            images (2d arrays). If a tuple is given it serves as the offset of
            the fitted area wrt the center of the 2d arrays.
        negative : bool, optional
            [method=2dfit/dftups/dftupspeckles] If True a negative 2d
            Gaussian/Moffat fit is performed.
        threshold : bool, optional
            [method=2dfit] If True the background pixels
            (estimated using sigma clipped statistics) will be replaced by
            small random Gaussian noise.
        save_shifts : bool, optional
            [method=2dfit/dftups] Whether to save the shifts to a file in disk.
        cy_1, cx_1 : int, optional
            [method=dftups] Coordinates of the center of the
            subimage for fitting a 2d Gaussian and centroiding the 1st frame.
        upsample_factor : int, optional
            [method=dftups] Upsampling factor (default 100).
            Images will be registered to within 1/upsample_factor of a pixel.
        alignment_iter : int, optional
            [method=dftupspeckles] Number of alignment
            iterations (recomputes median after each iteration).
        gamma : int, optional
            [method=dftupspeckles] Applies a gamma correction
            to emphasize speckles (useful for faint stars).
        min_spat_freq : float, optional
            [method=dftupspeckles] Spatial frequency for high
            pass filter.
        max_spat_freq : float, optional
            [method=dftupspeckles] Spatial frequency for low
            pass filter.
        recenter_median : bool, optional
            [method=dftupspeckles] Recenter the frames at each
            iteration based on the gaussian fit.
        sigfactor : int, optional
            [method=satspots] The background pixels will
            be thresholded before fitting a 2d Gaussian to the data using sigma
            clipped statistics. All values smaller than (MEDIAN +
            sigfactor*STDDEV) will be replaced by small random Gaussian noise.
        cropsize : odd int, optional
            [method=radon] Size in pixels of the cropped central area of the
            input array that will be used. It should be large enough to contain
            the satellite spots.
        hsize : float, optional
            [method=radon] Size of the box for the grid search. The frame is
            shifted to each direction from the center in a hsize length with a
            given step.
        step : float, optional
            [method=radon] The step of the coordinates change.
        mask_center : None or int, optional
            [method=radon] If None the central area of the frame is kept. If int
            a centered zero mask will be applied to the frame. By default the
            center isn't masked.
        verbose : bool, optional
            Whether to print to stdout the timing and aditional info.
        debug : bool, optional
            If True debug information is printed and plotted.
        plot : bool, optional
            Whether to plot the shifts.

        """

        if method == '2dfit':
            if self.fwhm is None:
                raise ValueError('FWHM has not been set')
            self.cube = cube_recenter_2dfit(
                self.cube, xy, self.fwhm, subi_size, model, nproc, imlib,
                interpolation, offset, negative, threshold, save_shifts, False,
                verbose, debug, plot
            )
        elif method == 'dftups':
            if self.fwhm is None:
                raise ValueError('FWHM has not been set')
            self.cube = cube_recenter_dft_upsampling(
                self.cube, cy_1, cx_1, negative, self.fwhm, subi_size,
                upsample_factor, imlib, interpolation, False, verbose,
                save_shifts, debug, plot
            )
        elif method == 'dftupspeckles':
            if self.fwhm is None:
                raise ValueError('FWHM has not been set')
            res = cube_recenter_via_speckles(
                self.cube, self.cuberef, alignment_iter, gamma, min_spat_freq,
                max_spat_freq, self.fwhm, debug, negative, recenter_median,
                subi_size, imlib, interpolation, plot
            )
            if self.cuberef is None:
                self.cube = res[0]
            else:
                self.cube = res[0]
                self.cuberef = res[1]
        elif method == 'satspots':
            self.cube, _, _ = cube_recenter_satspots(
                self.cube, xy, subi_size, sigfactor, plot, debug, verbose
            )
        elif method == 'radon':
            self.cube = cube_recenter_radon(
                self.cube, full_output=False, verbose=verbose, imlib=imlib,
                interpolation=interpolation, cropsize=cropsize, hsize=hsize,
                step=step, mask_center=mask_center, nproc=nproc, debug=debug
            )
        else:
            raise ValueError('Method not recognized')

    def remove_badframes(self, method='corr', frame_ref=None, crop_size=30,
                         dist='pearson', percentile=20, stat_region='annulus',
                         inner_radius=10, width=10, top_sigma=1.0,
                         low_sigma=1.0, window=None, roundlo=-0.2, roundhi=0.2,
                         lambda_ref=0, plot=True, verbose=True):
        """
        Find outlying/bad frames and slice the cube accordingly.

        Besides modifying ``self.cube`` and ``self.angles``, also sets a
        ``self.good_indices`` which contain the indices of the angles which were
        kept.

        Parameters
        ----------
        method : {'corr', 'pxstats', 'ellip'}, optional
            Method which is used to determine bad frames. Refer to the
            ``preproc.badframes`` submodule for explanation of the different
            methods.
        frame_ref : int, 2d array or None, optional
            [method=corr] Index of the frame that will be used as a reference or
            2d reference array.
        crop_size : int, optional
            [method=corr] Size in pixels of the square subframe to be analyzed.
        dist : {'sad','euclidean','mse','pearson','spearman'}, optional
            [method=corr] One of the similarity or dissimilarity measures from
            function vip_hci.stats.distances.cube_distance().
        percentile : float, optional
            [method=corr] The percentage of frames that will be discarded
            [0..100].
        stat_region : {'annulus', 'circle'}, optional
            [method=pxstats] Whether to take the statistics from a circle or an
            annulus.
        inner_radius : int, optional
            [method=pxstats] If stat_region is 'annulus' then 'in_radius' is the
            inner radius of the annular region. If stat_region is 'circle' then
            'in_radius' is the radius of the aperture.
        width : int, optional
            [method=pxstats] Size of the annulus. Ignored if mode is 'circle'.
        top_sigma : int, optional
            [method=pxstats] Top boundary for rejection.
        low_sigma : int, optional
            [method=pxstats] Lower boundary for rejection.
        window : int, optional
            [method=pxstats] Window for smoothing the median and getting the
            rejection statistic.
        roundlo,roundhi : float, optional
            [method=ellip] : Lower and higher bounds for the ellipticipy.
        lambda_ref : int, optional
            [4D cube] Which wavelength to consider when determining bad frames
            on a 4D cube.
        plot : bool, optional
            If true it plots the mean fluctuation as a function of the frames
            and the boundaries.
        verbose : bool, optional
            Show debug output.

        """
        if self.cube.ndim == 4:
            tcube = self.cube[lambda_ref]
        else:
            tcube = self.cube

        if method == 'corr':
            if frame_ref is None:
                print("Correlation method selected but `frame_ref` is missing")
                print("Setting the 1st frame as the reference")
                frame_ref = 0

            self.good_indices, _ = cube_detect_badfr_correlation(tcube,
                                        frame_ref, crop_size, dist, percentile,
                                        plot, verbose)
        elif method == 'pxstats':
            self.good_indices, _ = cube_detect_badfr_pxstats(tcube, stat_region,
                                        inner_radius, width, top_sigma,
                                        low_sigma, window, plot, verbose)
        elif method == 'ellip':
            if self.cube.ndim == 4:
                fwhm = self.fwhm[lambda_ref]
            else:
                fwhm = self.fwhm
            self.good_indices, _ = cube_detect_badfr_ellipticity(tcube, fwhm,
                                        crop_size, roundlo, roundhi, plot,
                                        verbose)
        else:
            raise ValueError('Bad frames detection method not recognized')

        if self.cube.ndim == 4:
            self.cube = self.cube[:, self.good_indices]
        else:
            self.cube = self.cube[self.good_indices]

        if verbose:
            print("New cube shape: {}".format(self.cube.shape))

        if self.angles is not None:
            self.angles = self.angles[self.good_indices]
            if verbose:
                msg = "New parallactic angles vector shape: {}"
                print(msg.format(self.angles.shape))

    def rescale(self, scale, imlib='ndimage', interpolation='bicubic',
                verbose=True):
        """ Resampling the pixels (upscaling or downscaling the frames).

        Parameters
        ----------
        scale : int, float or tuple
            Scale factor for upsampling or downsampling the frames in the cube.
            If a tuple it corresponds to the scale along x and y.
        imlib : {'ndimage', 'opencv'}, str optional
            Library used for image transformations. ndimage is the default.
        interpolation : str, optional
            For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic',
            'bicubic', 'biquartic', 'biquintic'. The 'nearneig' interpolation
            is the fastest and the 'biquintic' the slowest. The 'nearneig' is
            the worst option for interpolation of noisy astronomical images.
            For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
            The 'nearneig' interpolation is the fastest and the 'lanczos4' the
            slowest and accurate.
        verbose : bool, optional
            Whether to print out additional info such as the new cube shape.

        """
        self.cube = cube_px_resampling(self.cube, scale, imlib, interpolation,
                                       verbose)

    def subsample(self, window, mode='mean'):
        """ Temporally sub-sampling the sequence (3d or 4d cube).

        Parameters
        ----------
        window : int
            Window for mean/median.
        mode : {'mean', 'median'}, optional
            Switch for choosing mean or median.
        """
        if self.angles is not None:
            self.cube, self.angles = cube_subsample(self.cube, window,
                                                    mode, self.angles)
        else:
            self.cube = cube_subsample(self.cube, window, mode)



"""
Subpackage ``medsub`` has the ADI cube basic processing (Marois et al. 2006): median
frame subtraction, and annular mode where ``n`` closest frames taking into account
a PA threshold are median collapsed and subtracted.
"""

from .medsub_source import *


#! /usr/bin/env python

"""
Module with NMF algorithm for ADI.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez, Valentin Christiaens'
__all__ = ['nmf']

import numpy as np
from sklearn.decomposition import NMF
from ..preproc import cube_derotate, cube_collapse
from ..preproc.derotation import _compute_pa_thresh, _find_indices_adi
from ..var import (prepare_matrix, reshape_matrix, frame_center, dist, 
                   matrix_scaling, mask_circle)
from ..conf import timing, time_ini


def nmf(cube, angle_list, cube_ref=None, ncomp=1, scaling=None, max_iter=10000,
        random_state=None, mask_center_px=None, source_xy=None, delta_rot=1, 
        fwhm=4, init_svd='nndsvd', imlib='opencv', interpolation='lanczos4', 
        collapse='median', full_output=False, verbose=True, cube_sig=None, 
        handle_neg='mask', **kwargs):
    """ Non Negative Matrix Factorization for ADI sequences. Alternative to the
    full-frame ADI-PCA processing that does not rely on SVD or ED for obtaining
    a low-rank approximation of the datacube. This function embeds the 
    scikit-learn NMF algorithm solved through either the coordinate descent or 
    the multiplicative update method.
    
    Parameters
    ----------
    cube : numpy ndarray, 3d
        Input cube.
    angle_list : numpy ndarray, 1d
        Corresponding parallactic angle for each frame.   
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    ncomp : int, optional
        How many components are used as for low-rank approximation of the 
        datacube.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    max_iter : int optional
        The number of iterations for the coordinate descent solver.
    random_state : int or None, optional
        Controls the seed for the Pseudo Random Number generator.
    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.
    source_xy : tuple of int, optional
        For ADI-PCA, this triggers a frame rejection in the PCA library, with
        ``source_xy`` as the coordinates X,Y of the center of the annulus where
        the PA criterion is estimated. When ``ncomp`` is a tuple, a PCA grid is
        computed and the S/Ns (mean value in a 1xFWHM circular aperture) of the
        given (X,Y) coordinates are computed.
    delta_rot : int, optional
        Factor for tunning the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1xFHWM on each side of the considered frame).
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default value is 4. 
    init_svd: str, optional {'nnsvd','nnsvda','random'}
        Method used to initialize the iterative procedure to find H and W.
        'nndsvd': non-negative double SVD recommended for sparseness
        'nndsvda': NNDSVD where zeros are filled with the average of cube; 
        recommended when sparsity is not desired
        'random': random initial non-negative matrix
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    handle_neg: str, opt {'subtr_min','mask','null'}
        Determines how to handle negative values: mask them, set them to zero,
        or subtract the minimum value in the arrays. Note: 'mask' or 'null' 
        may leave significant artefacts after derotation of residual cube
        => those options should be used carefully (e.g. with proper treatment
        of masked values in non-derotated cube of residuals).
    kwargs 
        Additional arguments for scikit-learn NMF algorithm. See:
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html             
             
    Returns
    -------
    If full_output is False the final frame is returned. If True the algorithm
    returns the reshaped NMF components, the reconstructed cube, the residuals,
    the residuals derotated and the final frame.     
    
    """
    array = cube.copy()
    if verbose:  
        start_time = time_ini()
    n, y, x = array.shape
    
    matrix_ref=None
    matrix_sig=None
    
    # how to handle negative values
    if handle_neg == 'mask':
        if mask_center_px:
            array = mask_circle(array, mask_center_px)
        if cube_sig is not None:
            yy, xx = np.where(np.amin(array-np.abs(cube_sig),axis=0)>0)
        else:
            yy, xx = np.where(np.amin(array,axis=0)>0)
        H_tmp = np.zeros([ncomp,y,x])
        if len(yy)>0:
            matrix = array[:,yy,xx]
            matrix = matrix_scaling(matrix, scaling)
            if cube_ref is not None:
                matrix_ref = cube_ref[:,yy,xx]
                matrix_ref = matrix_scaling(matrix_ref, scaling)
            if cube_sig is not None:
                matrix_sig = cube_sig[:,yy,xx]              
        else:
            raise ValueError("Remove frame(s) with negative values")
    else:
        if handle_neg=='null':
            if cube_sig is not None:
                array[np.where(array-cube_sig<0)]=0
                cube_sig[np.where(array-cube_sig<0)]=0
            else:
                array[np.where(array<0)]=0
            
        elif handle_neg=='subtr_min':
            if cube_sig is not None:
                array -= np.amin(array-cube_sig)
            else:
                array -= np.amin(array)
        else:
            raise ValueError("Mode to handle neg. pixels not recognized")
    
        matrix = prepare_matrix(array, scaling, mask_center_px, mode='fullfr',
                                verbose=verbose)
        if cube_ref is not None:
            matrix_ref = prepare_matrix(cube_ref, scaling, mask_center_px,
                                        mode='fullfr', verbose=verbose)
        if cube_sig is not None:
            matrix_sig = prepare_matrix(cube_sig, scaling, mask_center_px, 
                                        mode='fullfr', verbose=verbose)
   

    # elif source_xy is None or delta_rot is None:
    #     pass
    # else:
    #     # rotation threshold applied and NMF repeated for each frame on different matrices
    #     if delta_rot is None or fwhm is None:
    #         msg = 'Delta_rot or fwhm parameters missing. Needed for the'
    #         msg += 'PA-based rejection of frames from the library'
    #         raise TypeError(msg)
    #     residuals_cube = np.zeros_like(cube)
    #     recon_cube = np.zeros_like(cube)
    #     yc, xc = frame_center(cube[0], False)
    #     x1, y1 = source_xy
    #     ann_center = dist(yc, xc, y1, x1)
    #     pa_thr = _compute_pa_thresh(ann_center, fwhm, delta_rot)
    #     mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list))/2
    #     if pa_thr >= mid_range - mid_range * 0.1:
    #         new_pa_th = float(mid_range - mid_range * 0.1)
    #         if verbose:
    #             msg = 'PA threshold {:.2f} is too big, will be set to '
    #             msg += '{:.2f}'
    #             print(msg.format(pa_thr, new_pa_th))
    #         pa_thr = new_pa_th

    #     for frame in range(n):
    #         if ann_center > fwhm * 3:  # TODO: 3 optimal value? new par?
    #             ind = _find_indices_adi(angle_list, frame, pa_thr,
    #                                     truncate=True)
    #         else:
    #             ind = _find_indices_adi(angle_list, frame, pa_thr)
    

    if cube_sig is not None:
        # derotate
        residuals_cube = cube_sig.copy()
    else:
        residuals_cube = np.zeros_like(array)
    
    
    if source_xy is None:
        residuals = _project_subtract(matrix, matrix_ref, ncomp, scaling,
                                      mask_center_px, verbose, full_output, 
                                      matrix_sig=matrix_sig, max_iter=max_iter, 
                                      random_state=random_state, 
                                      init_svd=init_svd, **kwargs)
        if verbose:
            timing(start_time)
        if full_output:
            #reconstructed = residuals[1]
            H = residuals[2]
            reconstructed = residuals[1]
            residuals = residuals[0]
        recon_cube = residuals_cube.copy()
        if handle_neg=='mask':
            for fr in range(n):
                residuals_cube[fr][yy, xx] = residuals[fr]
            if full_output:
                for fr in range(n):
                    recon_cube[fr][yy, xx] = reconstructed[fr]
                for pp in range(ncomp):
                    H_tmp[pp][yy,xx] = H[pp]
                H = H_tmp
        else:
            for fr in range(n):
                residuals_cube[fr] = residuals[fr].reshape((y, x))
            if full_output:
                recon_cube = reshape_matrix(reconstructed, y,x)
                H = H.reshape(ncomp,y,x)
    else:
        if delta_rot is None or fwhm is None:
            msg = 'Delta_rot or fwhm parameters missing. Needed for the'
            msg += 'PA-based rejection of frames from the library'
            raise TypeError(msg)
        recon_cube = np.zeros_like(cube)
        yc, xc = frame_center(cube[0], False)
        x1, y1 = source_xy
        ann_center = dist(yc, xc, y1, x1)
        pa_thr = _compute_pa_thresh(ann_center, fwhm, delta_rot)
        mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list))/2
        if pa_thr >= mid_range - mid_range * 0.1:
            new_pa_th = float(mid_range - mid_range * 0.1)
            if verbose:
                msg = 'PA threshold {:.2f} is too big, will be set to '
                msg += '{:.2f}'
                print(msg.format(pa_thr, new_pa_th))
            pa_thr = new_pa_th

        for fr in range(n):
            ind = _find_indices_adi(angle_list, fr, pa_thr)
            res_result = _project_subtract(matrix, matrix_ref, ncomp, scaling,
                                           mask_center_px, verbose, 
                                           full_output, ind, fr, 
                                           matrix_sig=matrix_sig, 
                                           max_iter=max_iter, 
                                           random_state=random_state, 
                                           init_svd=init_svd, **kwargs)
            # ! Instead of reshaping, fill frame using get_annulus?
            if full_output:
                residuals = res_result[0]
                recon_frame = res_result[1]
                H = res_result[2]
                if handle_neg=='mask':
                    recon_cube[fr][yy,xx] = recon_frame
                else:
                    recon_cube[fr] = recon_frame.reshape((y, x))
            else:
                residuals = res_result
            if handle_neg=='mask':
                residuals_cube[fr][yy, xx] = residuals
                if fr == n-1:
                    for pp in range(ncomp):
                        H_tmp[pp][yy, xx] = H[pp]
                    H = H_tmp
            else:
                residuals_cube[fr] = residuals.reshape((y, x))
                if fr == n-1:
                    H = H.reshape(ncomp,y,x)
                
    if verbose:  
        print('Done NMF with sklearn.NMF.')
        timing(start_time)
            
    residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                    interpolation=interpolation)
    frame = cube_collapse(residuals_cube_, mode=collapse)
    
    if verbose:  
        print('Done derotating and combining.')
        timing(start_time)
    if full_output:
        return (H, recon_cube, residuals_cube,
                residuals_cube_, frame)
    else:
        return frame
    
    
def _project_subtract(matrix, matrix_ref, ncomp, scaling, mask_center_px, 
                      verbose, full_output, indices=None, frame=None, 
                      matrix_sig=None, max_iter=100, random_state=None, 
                      init_svd='nndsvd', **kwargs):
    """
    PCA projection and model PSF subtraction. Used as a helping function by
    each of the PCA modes (ADI, ADI+RDI, ADI+mSDI).

    Parameters
    ----------
    cube : numpy ndarray
        Input cube.
    cube_ref : numpy ndarray
        Reference cube.
    ncomp : int
        Number of principal components.
    scaling : str
        Scaling of pixel values. See ``pca`` docstrings.
    mask_center_px : int
        Masking out a centered circular aperture.
    verbose : bool
        Verbosity.
    full_output : bool
        Whether to return intermediate arrays or not.
    indices : list
        Indices to be used to discard frames (a rotation threshold is used).
    frame : int
        Index of the current frame (when indices is a list and a rotation
        threshold was applied).
    cube_sig: numpy ndarray, opt
        Cube with estimate of significant authentic signals. If provided, this
        will be subtracted from both the cube and the PCA library, before 
        projecting the cube onto the principal components.
        
    Returns
    -------
    ref_lib_shape : int
        [indices is not None, frame is not None] Number of
        rows in the reference library for the given frame.
    residuals: numpy ndarray
        Residuals, returned in every case.
    reconstructed : numpy ndarray
        [full_output=True] The reconstructed array.
    V : numpy ndarray
        [full_output=True, indices is None, frame is None] The right singular
        vectors of the input matrix, as returned by ``svd/svd_wrapper()``
    """    
    if matrix_sig is None:
        matrix_emp = matrix.copy()
    else:
        matrix_emp = matrix-matrix_sig
        
    if matrix_ref is not None:
        ref_lib = matrix_ref
    elif indices is not None and frame is not None:
        ref_lib = matrix_emp[indices].copy()
    else:
        ref_lib = matrix_emp.copy()

    # to avoid bug, just consider positive values
    if np.median(ref_lib)<0:
        raise ValueError("Mostly negative values in the cube")
    else:
        ref_lib[np.where(ref_lib<0)] = 0

    solver = 'mu'
    # if init_svd != 'nndsvd':
    #     solver = 'mu'
    # else:
    #     solver = 'cd'
    mod = NMF(n_components=ncomp, alpha=0, solver=solver, init=init_svd, 
              max_iter=max_iter, random_state=random_state, **kwargs)   
    # a rotation threshold is used (frames are processed one by one)
    if indices is not None and frame is not None:
        if ref_lib.shape[0] <= 10:
            raise RuntimeError('Less than 10 frames left in the PCA library'
                               ', Try decreasing the parameter delta_rot')
        curr_frame = matrix[frame]  # current frame
        curr_frame_emp = matrix_emp[frame]
        # H [ncomp, n_pixels]
        H = mod.fit(ref_lib).components_
        # W: coefficients [1, ncomp]
        W = mod.transform(curr_frame_emp[np.newaxis,...])
        #V = svd_wrapper(ref_lib, svd_mode, ncomp, False)
        #transformed = np.dot(curr_frame_emp, V.T)
        #reconstructed = np.dot(transformed.T, V)
        reconstructed = np.dot(W, H)
        residuals = curr_frame - reconstructed
        if full_output:
            return residuals, reconstructed, H
        else:
            return residuals

    # the whole matrix is processed at once
    else:      
        # H [ncomp, n_pixels]: Non-negative components of the data
        #if cube_ref is not None:
        H = mod.fit(ref_lib).components_
        #else:
        #    H = mod.fit(matrix).components_          
        
        # W: coefficients [n_frames, ncomp]
        W = mod.transform(matrix_emp)
            
        reconstructed = np.dot(W, H)
        residuals = matrix - reconstructed
        
        # V = svd_wrapper(ref_lib, svd_mode, ncomp, verbose)
        # transformed = np.dot(V, matrix_emp.T)
        # reconstructed = np.dot(transformed.T, V)
        # residuals = matrix - reconstructed
        # residuals_res = reshape_matrix(residuals, y, x)
        if full_output:
            return residuals, reconstructed, H
        else:
            return residuals


#! /usr/bin/env python

"""
Module with sigma clipping functions.
"""



__author__ = 'Carlos Alberto Gomez Gonzalez', 'V. Christiaens'
__all__ = ['clip_array',
           'sigma_filter']

import numpy as np

from scipy.ndimage.filters import generic_filter
from astropy.stats import median_absolute_deviation
import warnings
try:
    from numba import njit
    no_numba = False
except ImportError:
    msg = "Numba python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_numba = True


def sigma_filter(frame_tmp, bpix_map, neighbor_box=3, min_neighbors=3, 
                 verbose=False):
    """Sigma filtering of pixels in a 2d array.
    
    Parameters
    ----------
    frame_tmp : numpy ndarray
        Input 2d array, image.
    bpix_map: numpy ndarray
        Input array of the same size as frame_tmp, indicating the locations of
        bad/nan pixels by 1 (the rest of the array is set to 0)
    neighbor_box : int, optional
        The side of the square window around each pixel where the sigma and
        median are calculated.
    min_neighbors : int, optional
        Minimum number of good neighboring pixels to be able to correct the
        bad/nan pixels
        
    Returns
    -------
    frame_corr : numpy ndarray
        Output array with corrected bad/nan pixels
    
    """
    if not no_numba:
        @njit
        def _sigma_filter_numba(frame_tmp, bpix_map, neighbor_box=3, 
                                min_neighbors=3, verbose=False):
            if not frame_tmp.ndim == 2:
                raise TypeError('Input array is not a frame or 2d array')
        
            sz_y = frame_tmp.shape[0]  # get image y-dim
            sz_x = frame_tmp.shape[1]  # get image x-dim
            bp = bpix_map.copy()       # temporary bpix map; important to make a copy!
            im = frame_tmp             # corrected image
            nb = int(np.sum(bpix_map)) # number of bad pixels remaining
            nit = 0                                 # number of iterations
            while nb > 0:
                nit +=1
                wb = np.where(bp)                   # find bad pixels
                gp = 1 - bp                         # temporary good pixel map
                for n in range(nb):
                    #0/ Determine the box around each pixel
                    half_box = int(np.floor(neighbor_box/2.))
                    hbox_b = min(half_box,wb[0][n])        # half size of the box at
                                                           # the bottom of the pixel
                    hbox_t = min(half_box,sz_y-1-wb[0][n]) # half size of the box at 
                                                           # the top of the pixel
                    hbox_l = min(half_box,wb[1][n])        # half size of the box to 
                                                           # the left of the pixel
                    hbox_r = min(half_box,sz_x-1-wb[1][n]) # half size of the box to 
                                                           # the right of the pixel
                    # in case we are close to an edge, we want to extend the box
                    # in the direction opposite to the edge: 
                    if hbox_b<hbox_t: 
                        hbox_t += half_box-hbox_b
                    elif hbox_t<hbox_b: 
                        hbox_b += half_box-hbox_t
                    if hbox_l<hbox_r: 
                        hbox_r += half_box-hbox_l
                    elif hbox_r<hbox_l:
                        hbox_l += half_box-hbox_r
                    sgp = gp[(wb[0][n]-hbox_b):(wb[0][n]+hbox_t+1),
                             (wb[1][n]-hbox_l):(wb[1][n]+hbox_r+1)]
                    if int(np.sum(sgp)) >= min_neighbors:
                        sim = im[(wb[0][n]-hbox_b):(wb[0][n]+hbox_t+1),
                                 (wb[1][n]-hbox_l):(wb[1][n]+hbox_r+1)]
                        px_x = int(wb[0][n])
                        px_y = int(wb[1][n])
                        gsgp = np.where(sgp)
                        gsim=[]
                        for i in range(len(gsgp[0])):
                            gsim.append(sim[gsgp[0][i],gsgp[1][i]])
                        im[px_x,px_y] = np.median(np.array(gsim))
                        bp[px_x,px_y] = 0
                nb = int(np.sum(bp))
            if verbose:
                print('Required number of iterations in the sigma filter: ', nit)
    
            return im

    # TODO: If possible, replace this function using
    # scipy.ndimage.filters.generic_filter and astropy.stats.sigma_clip
    def _sigma_filter(frame_tmp, bpix_map, neighbor_box=3, min_neighbors=3,
                     verbose=False):
        """Same description as wrapper function."""
        
        if frame_tmp.ndim != 2:
            raise TypeError('Input array is not a frame or 2d array')
    
        sz_y = frame_tmp.shape[0]  # get image y-dim
        sz_x = frame_tmp.shape[1]  # get image x-dim
        bp = bpix_map.copy()       # temporary bpix map; important to make a copy!
        im = frame_tmp             # corrected image
        nb = int(np.sum(bpix_map)) # number of bad pixels remaining
        #In each iteration, correct only the bpix with sufficient good 'neighbors'
        nit = 0                                 # number of iterations
        while nb > 0:
            nit += 1
            wb = np.where(bp)                   # find bad pixels
            gp = 1 - bp                         # temporary good pixel map
            for n in range(nb):
                #0/ Determine the box around each pixel
                half_box = np.floor(neighbor_box/2)
                hbox_b = min(half_box, wb[0][n])       # half size of the box at the
                                                       # bottom of the pixel
                hbox_t = min(half_box, sz_y-1-wb[0][n])# half size of the box at the
                                                       # top of the pixel
                hbox_l = min(half_box, wb[1][n])       # half size of the box to the
                                                       # left of the pixel
                hbox_r = min(half_box, sz_x-1-wb[1][n])# half size of the box to the
                                                       # right of the pixel
                # but in case we are at an edge, we want to extend the box by one 
                # row/column of pixels in the direction opposite to the edge to 
                # have 9 px instead of 6: 
                if half_box == 1:
                    if wb[0][n] == sz_y-1:
                        hbox_b = hbox_b+1
                    elif wb[0][n] == 0:
                        hbox_t = hbox_t+1
                    if wb[1][n] == sz_x-1:
                        hbox_l = hbox_l+1
                    elif wb[1][n] == 0:
                        hbox_r = hbox_r+1
    
                sgp = gp[int(wb[0][n]-hbox_b): int(wb[0][n]+hbox_t+1),
                         int(wb[1][n]-hbox_l): int(wb[1][n]+hbox_r+1)]
                if int(np.sum(sgp)) >= min_neighbors:
                    sim = im[int(wb[0][n]-hbox_b): int(wb[0][n]+hbox_t+1),
                             int(wb[1][n]-hbox_l): int(wb[1][n]+hbox_r+1)]
                    im[wb[0][n],wb[1][n]] = np.median(sim[np.where(sgp)])
                    bp[wb[0][n],wb[1][n]] = 0
            nb = int(np.sum(bp))
        if verbose:
            print('Required number of iterations in the sigma filter: ', nit)
        return im


    if no_numba:
        return _sigma_filter(frame_tmp, bpix_map, neighbor_box=3, 
                             min_neighbors=3, verbose=False)
    else:
        return _sigma_filter_numba(frame_tmp, bpix_map, neighbor_box=3, 
                                   min_neighbors=3, verbose=False)



def clip_array(array, lower_sigma, upper_sigma, out_good=False, neighbor=False,
               num_neighbor=3, mad=False):
    """Sigma clipping for detecting outlying values in 2d array. If the
    parameter 'neighbor' is True the clipping can be performed in a local patch
    around each pixel, whose size depends on 'neighbor' parameter.
    
    Parameters
    ----------
    array : numpy ndarray
        Input 2d array, image.
    lower_sigma : float 
        Value for sigma, lower boundary.
    upper_sigma : float 
        Value for sigma, upper boundary.
    out_good : bool, optional
        For choosing different outputs.
    neighbor : bool, optional
        For clipping over the median of the contiguous pixels.
    num_neighbor : int, optional
        The side of the square window around each pixel where the sigma and 
        median are calculated. 
    mad : {False, True}, bool optional
        If True, the median absolute deviation will be used instead of the 
        standard deviation.
    min_std : float, optional
        Min standard deviation considered for the lower and upper sigma 
        conditions. Can be useful for frames with padded edges (e.g. SPHERE/IFS)
        
    Returns
    -------
    good : array_like
        If out_good argument is true, returns the indices of not-outlying px.
    bad : array_like 
        If out_good argument is false, returns a vector with the outlier px.
    
    """
    
    if not no_numba:
        @njit
        def _clip_array_numba(array, lower_sigma, upper_sigma, out_good=False, 
                              neighbor=False, num_neighbor=3, mad=False): 
            if array.ndim != 2:
                raise TypeError("Input array is not two dimensional (frame)\n")
                
            ny, nx = array.shape
            bpm = array.copy()    
            gpm = array.copy()
            
            if neighbor and num_neighbor:
                for y in range(ny):
                    for x in range(nx):
                        #0/ Determine the box around each pixel
                        half_box = int(np.floor(num_neighbor/2.))
                        hbox_b = min(half_box,y)    # half size of the box at the
                                                    # bottom of the pixel
                        hbox_t = min(half_box,ny-y) # half size of the box at the
                                                    # top of the pixel
                        hbox_l = min(half_box,x)    # half size of the box to the
                                                    # left of the pixel
                        hbox_r = min(half_box,nx-x) # half size of the box to the
                                                    # right of the pixel
                        # in case we are close to an edge, we want to extend the box
                        # in the direction opposite to the edge: 
                        if hbox_b<hbox_t: 
                            hbox_t += half_box-hbox_b
                        elif hbox_t<hbox_b: 
                            hbox_b += half_box-hbox_t
                        if hbox_l<hbox_r: 
                            hbox_r += half_box-hbox_l
                        elif hbox_r<hbox_l:
                            hbox_l += half_box-hbox_r
                            
                        sub_arr = array[y-hbox_b:y+hbox_t+1,
                                        x-hbox_l:x+hbox_r+1]
                        neighbours = sub_arr.flatten()
                        neigh_list = []
                        for i in range(neighbours.shape[0]):
                            neigh_list.append(neighbours[i])
                        neigh_list.remove(array[y,x])
                        
                        neigh_arr = np.array(neigh_list)
                        median=np.median(neigh_arr)
                        if mad:
                            abs_diff = []
                            for i in range(num_neighbor*num_neighbor-1):
                                abs_diff.append(np.absolute(median-neigh_arr[i]))
                            sigma = np.median(np.array(abs_diff))
                        else:
                            sigma = np.std(neigh_arr)
                        bad1 = array[y,x] < (median - lower_sigma * sigma) 
                        bad2 = array[y,x] > (median + upper_sigma * sigma)
                        bpm[y,x] = bad1 | bad2
                        gpm[y,x] = 1.-bpm[y,x]
            else:
                median = np.median(array)
                sigma = np.std(array)
                for y in range(ny):
                    for x in range(nx):
                        bad1 = array[y,x] < (median - lower_sigma * sigma) 
                        bad2 = array[y,x] > (median + upper_sigma * sigma)
                        bpm[y,x] = bad1 | bad2   
                        gpm[y,x] = 1.-bpm[y,x]
            
            if out_good:
                good = np.where(gpm)
                return good
            else:
                bad = np.where(bpm)
                return bad
            
    # TODO: If possible, replace this function using astropy.stats.sigma_clip
    def _clip_array(array, lower_sigma, upper_sigma, out_good=False, neighbor=False,
                    num_neighbor=3, mad=False):
        """Sigma clipping for detecting outlying values in 2d array. If the
        parameter 'neighbor' is True the clipping can be performed in a local patch
        around each pixel, whose size depends on 'neighbor' parameter.
        
        Parameters
        ----------
        array : numpy ndarray 
            Input 2d array, image.
        lower_sigma : float 
            Value for sigma, lower boundary.
        upper_sigma : float 
            Value for sigma, upper boundary.
        out_good : bool, optional
            For choosing different outputs.
        neighbor : bool optional
            For clipping over the median of the contiguous pixels.
        num_neighbor : int, optional
            The side of the square window around each pixel where the sigma and 
            median are calculated. 
        mad : bool, optional
            If True, the median absolute deviation will be used instead of the 
            standard deviation.
            
        Returns
        -------
        good : numpy ndarray
            If out_good argument is true, returns the indices of not-outlying px.
        bad : numpy ndarray 
            If out_good argument is false, returns a vector with the outlier px.
        
        """
        if array.ndim != 2:
            raise TypeError("Input array is not two dimensional (frame)\n")
    
        values = array.copy()
        if neighbor and num_neighbor:
            median = generic_filter(array, function=np.median, 
                                    size=(num_neighbor,num_neighbor), mode="mirror")
            if mad:
                sigma = generic_filter(array, function=median_absolute_deviation,                                 
                                       size=(num_neighbor,num_neighbor), 
                                       mode="mirror")
            else:
                sigma = generic_filter(array, function=np.std,                                 
                                       size=(num_neighbor,num_neighbor), 
                                       mode="mirror")
        else:
            median = np.median(values)
            sigma = values.std()
            
        good1 = values > (median - lower_sigma * sigma) 
        good2 = values < (median + upper_sigma * sigma)
        bad1 = values < (median - lower_sigma * sigma)
        bad2 = values > (median + upper_sigma * sigma)
        
        if out_good:
            # normal px indices in both good1 and good2
            good = np.where(good1 & good2)
            return good
        else:
            # deviating px indices in either bad1 or bad2
            bad = np.where(bad1 | bad2)
            return bad
        


    if no_numba:
        return _clip_array(array, lower_sigma, upper_sigma, out_good=out_good, 
                           neighbor=neighbor, num_neighbor=num_neighbor, 
                           mad=mad)
    else:
        return _clip_array_numba(array, lower_sigma, upper_sigma, 
                                 out_good=out_good, neighbor=neighbor, 
                                 num_neighbor=num_neighbor, mad=mad)


"""
Util functions for ANDROMEDA.
"""


__author__ = 'Ralf Farkas'
__all__ = []


import numpy as np


def robust_std(x):
    """
    Calculate and return the *robust* standard deviation of a point set.

    Corresponds to the *standard deviation* of the point set, without taking
    into account the outlier points.

    Parameters
    ----------
    x : numpy ndarray
        point set

    Returns
    -------
    std : np.float64
        robust standard deviation of the point set

    Notes
    -----
    based on ``LibAndromeda/robust_stddev.pro`` v1.1 2016/02/16

    """
    median_absolute_deviation = np.median(np.abs(x - np.median(x)))
    return median_absolute_deviation / 0.6745


def idl_round(x):
    """
    Round to the *nearest* integer, half-away-from-zero.

    Parameters
    ----------
    x : array-like
        Number or array to be rounded

    Returns
    -------
    r_rounded : array-like
        note that the returned values are floats

    Notes
    -----
    IDL ``ROUND`` rounds to the *nearest* integer (commercial rounding),
    unlike numpy's round/rint, which round to the nearest *even*
    value (half-to-even, financial rounding) as defined in IEEE-754
    standard.

    """
    return np.trunc(x + np.copysign(0.5, x))


def idl_where(array_expression):
    """
    Return a list of indices matching the ``array_expression``.

    Port of IDL's ``WHERE`` function.

    Parameters
    ----------
    array_expression : numpy ndarray / expression
        an expression like ``array > 0``

    Returns
    -------
    res : ndarray
        list of 'good' indices


    Notes
    -----
    - The IDL version returns ``[-1]`` when no match was found, this function
      returns ``[]``, which is more "pythonic".

    """
    res = np.array([i for i, e in enumerate(array_expression.flatten()) if e])
    return res
